###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""A mock of the BookkeepingClient, used for testing purposes."""
# pylint: disable=protected-access, missing-docstring, invalid-name, line-too-long

from unittest.mock import MagicMock

bkc_mock = MagicMock()
bkc_mock.sendXMLBookkeepingReport.return_value = {"OK": True, "Value": ""}
bkc_mock.getFileTypes.return_value = {
    "OK": True,
    "rpcStub": (
        ("Bookkeeping/BookkeepingManager", {"skipCACheck": False, "delegatedGroup": "diracAdmin", "timeout": 3600}),
        "getFileTypes",
        ({},),
    ),
    "Value": {
        "TotalRecords": 48,
        "ParameterNames": ["FileTypes"],
        "Records": [
            ["DAVINCIHIST"],
            ["DIELECTRON.DST"],
            ["BU2JPSIK.MDST"],
            ["SIM"],
            ["BD2JPSIKS.MDST"],
            ["BU2JPSIK.DST"],
            ["BUBDBSSELECTION.DST"],
            ["LAMBDA.DST"],
            ["BSMUMUBLIND.DST"],
            ["HADRONIC.DST"],
        ],
    },
}
bkc_mock.getFileMetadata.return_value = {
    "OK": True,
    "Value": {
        "Successful": {
            "foo": {
                "ADLER32": None,
                "FileType": "SDST",
                "FullStat": None,
                "GotReplica": "Yes",
                "DataqualityFlag": "OK",
                "EventStat": 101,
                "EventType": 12365501,
                "FileSize": 1000,
                "GUID": "7EA19086-1E36-E611-9457-001E67398520",
                "RunNumber": 93718,
            },
            "bar": {
                "ADLER32": None,
                "FileType": "SDST",
                "FullStat": None,
                "GotReplica": "Yes",
                "DataqualityFlag": "OK",
                "EventStat": 102,
                "EventType": 12365501,
                "FileSize": 2000,
                "GUID": "7EA19086-1E36-E611-9457-001E67398520",
                "RunNumber": 93720,
            },
        }
    },
    "rpcStub": (("Bookkeeping/BookkeepingManager",)),
}
bkc_mock.getFileTypeVersion.return_value = {"OK": True, "Value": {"lfn1": "ROOT", "lfn2": "MDF"}}
bkc_mock.getListOfRunsInProd.return_value = {"OK": True, "Value": [97019]}


def fake_getInputOutputFilesForProd(prod, run_number):
    if prod != 0:
        raise NotImplementedError("Only prod 0 is implemented")
    if run_number != 97019:
        raise NotImplementedError("Only run 97019 is implemented")
    return {
        "OK": True,
        "Value": [
            ["aa.raw", "bb.raw", "Yes", "RAW"],
            ["aa.raw", "bb.log", "Yes", "LOG"],
        ],
    }


bkc_mock.getInputOutputFilesForProd.side_effect = fake_getInputOutputFilesForProd
bkc_mock.getFileDescendents.return_value = {
    "OK": True,
    "Value": {
        "Failed": [],
        "NotProcessed": [],
        "Successful": {"aa.raw": ["bb.raw", "bb.log"]},
        "WithMetadata": {
            "aa.raw": {
                "bb.raw": {"FileType": "RAW", "RunNumber": 97019, "GotReplica": "Yes"},
                "bb.log": {"FileType": "LOG", "GotReplica": "Yes"},
            }
        },
    },
}


class BookkeepingClientFake:
    """a fake BookkeepingClient - replicating some of the methods"""

    def getAvailableSteps(self, stepID):
        if stepID == {"StepId": 123}:
            return {
                "OK": True,
                "Value": {
                    "TotalRecords": 1,
                    "ParameterNames": [
                        "StepId",
                        "StepName",
                        "ApplicationName",
                        "ApplicationVersion",
                        "OptionFiles",
                        "Visible",
                        "ExtraPackages",
                        "ProcessingPass",
                        "OptionsFormat",
                        "DDDB",
                        "CONDDB",
                        "DQTag",
                        "SystemConfig",
                    ],
                    "Records": [
                        [
                            123,
                            "Stripping14-Stripping",
                            "DaVinci",
                            "v2r2",
                            "optsFiles",
                            "Yes",
                            "eps",
                            "procPass",
                            "",
                            "",
                            "123456",
                            "",
                            "",
                        ]
                    ],
                },
            }
        elif stepID in ({"StepId": 456}, {"StepId": 987}):
            return {
                "OK": True,
                "Value": {
                    "TotalRecords": 1,
                    "ParameterNames": [
                        "StepId",
                        "StepName",
                        "ApplicationName",
                        "ApplicationVersion",
                        "OptionFiles",
                        "Visible",
                        "ExtraPackages",
                        "ProcessingPass",
                        "OptionsFormat",
                        "DDDB",
                        "CONDDB",
                        "DQTag",
                        "SystemConfig",
                    ],
                    "Records": [
                        [
                            456,
                            "Merge",
                            "LHCb",
                            "v1r2",
                            "optsFiles",
                            "Yes",
                            "eps",
                            "procPass",
                            "",
                            "",
                            "fromPreviousStep",
                            "",
                            "x86_64-slc6-gcc49-opt",
                        ]
                    ],
                },
            }
        elif stepID == {"StepId": 789}:
            return {
                "OK": True,
                "Value": {
                    "TotalRecords": 1,
                    "ParameterNames": [
                        "StepId",
                        "StepName",
                        "ApplicationName",
                        "ApplicationVersion",
                        "OptionFiles",
                        "Visible",
                        "ExtraPackages",
                        "ProcessingPass",
                        "OptionsFormat",
                        "DDDB",
                        "CONDDB",
                        "DQTag",
                        "SystemConfig",
                    ],
                    "Records": [
                        [
                            789,
                            "MergeHisto",
                            "Noether",
                            "v1r2",
                            "optsFiles",
                            "Yes",
                            "eps",
                            "procPass",
                            "",
                            "",
                            "fromPreviousStep",
                            "",
                            "x86_64-slc6-gcc49-opt",
                        ]
                    ],
                },
            }
        elif stepID == {"StepId": 125080}:
            options125080 = "$APPCONFIGOPTS/Gauss/Sim08-Beam4000GeV-mu100-2012-nu2.5.py;"
            options125080 += "$DECFILESROOT/options/11102400.py;"
            options125080 += "$LBPYTHIA8ROOT/options/Pythia8.py;"
            options125080 += "$APPCONFIGOPTS/Gauss/G4PL_FTFP_BERT_EmNoCuts.py;"
            options125080 += "$APPCONFIGOPTS/Persistency/Compression-ZLIB-1.py"
            return {
                "OK": True,
                "Value": {
                    "TotalRecords": 1,
                    "ParameterNames": [
                        "StepId",
                        "StepName",
                        "ApplicationName",
                        "ApplicationVersion",
                        "ExtraPackages",
                        "ProcessingPass",
                        "Visible",
                        "Usable",
                        "DDDB",
                        "CONDDB",
                        "DQTag",
                        "OptionsFormat",
                        "OptionFiles",
                        "isMulticore",
                        "SystemConfig",
                        "mcTCK",
                        "ExtraOptions",
                    ],
                    "Records": [
                        [
                            125080,
                            "Sim08a",
                            "Gauss",
                            "v45r3",
                            "AppConfig.v3r171",
                            "Sim08a",
                            "Y",
                            "Yes",
                            "Sim08-20130503-1",
                            "Sim08-20130503-1-vc-mu100",
                            "",
                            "",
                            options125080,
                            "N",
                            "x86_64-slc5-gcc43-opt",
                            "",
                            "",
                        ]
                    ],
                },
            }
        elif stepID == {"StepId": 124620}:
            options124620 = "$APPCONFIGOPTS/Boole/Default.py;"
            options124620 += "$APPCONFIGOPTS/Boole/DataType-2012.py;"
            options124620 += "$APPCONFIGOPTS/Persistency/Compression-ZLIB-1.py"
            return {
                "OK": True,
                "Value": {
                    "TotalRecords": 1,
                    "ParameterNames": [
                        "StepId",
                        "StepName",
                        "ApplicationName",
                        "ApplicationVersion",
                        "ExtraPackages",
                        "ProcessingPass",
                        "Visible",
                        "Usable",
                        "DDDB",
                        "CONDDB",
                        "DQTag",
                        "OptionsFormat",
                        "OptionFiles",
                        "isMulticore",
                        "SystemConfig",
                        "mcTCK",
                        "ExtraOptions",
                    ],
                    "Records": [
                        [
                            124620,
                            "Digi13",
                            "Boole",
                            "v26r3",
                            "AppConfig.v3r164",
                            "Digi13",
                            "N",
                            "Yes",
                            "Sim08-20130503-1",
                            "Sim08-20130503-1-vc-mu100",
                            "",
                            "",
                            options124620,
                            "N",
                            "x86_64-slc5-gcc43-opt",
                            "",
                            "",
                        ]
                    ],
                },
            }
        elif stepID == {"StepId": 124621}:
            return {
                "OK": True,
                "Value": {
                    "TotalRecords": 1,
                    "ParameterNames": [
                        "StepId",
                        "StepName",
                        "ApplicationName",
                        "ApplicationVersion",
                        "ExtraPackages",
                        "ProcessingPass",
                        "Visible",
                        "Usable",
                        "DDDB",
                        "CONDDB",
                        "DQTag",
                        "OptionsFormat",
                        "OptionFiles",
                        "isMulticore",
                        "SystemConfig",
                        "mcTCK",
                        "ExtraOptions",
                    ],
                    "Records": [
                        [
                            124621,
                            "Digi13",
                            "Brunel",
                            "v26r3",
                            "AppConfig.v3r164",
                            "Digi13",
                            "N",
                            "Yes",
                            "Sim08-20130503-1",
                            "Sim08-20130503-1-vc-mu100",
                            "",
                            "",
                            "someBrunelOptions",
                            "N",
                            "x86_64-slc5-gcc43-opt",
                            "",
                            "",
                        ]
                    ],
                },
            }
        elif stepID == {"StepId": 999}:
            return {
                "OK": True,
                "Value": {
                    "TotalRecords": 1,
                    "ParameterNames": [
                        "StepId",
                        "OptionFiles",
                        "ExtraOptions",
                        "SystemConfig",
                        "mcTCK",
                        "isMulticore",
                        "StepName",
                        "ApplicationName",
                        "ApplicationVersion",
                        "Visible",
                        "Usable",
                        "ProcessingPass",
                        "ExtraPackages",
                        "DDDB",
                        "CONDDB",
                        "DQTag",
                    ],
                    "Records": [
                        [
                            999,
                            "$APPCONFIGOPTS/Persistency/Compression-ZLIB-1.py",
                            "",
                            "x86_64-slc5-gcc49-opt",
                            "",
                            "N",
                            "Stripping28",
                            "DaVinci",
                            "v41r3",
                            "Yes",
                            "Yes",
                            "Stripping28",
                            "AppConfig.v3r306",
                            "dddb-20150724",
                            "cond-20161011",
                            "",
                        ]
                    ],
                },
            }
        elif stepID == {"StepId": 998}:
            return {
                "OK": True,
                "Value": {
                    "TotalRecords": 1,
                    "ParameterNames": [
                        "StepId",
                        "OptionFiles",
                        "ExtraOptions",
                        "SystemConfig",
                        "mcTCK",
                        "isMulticore",
                        "StepName",
                        "ApplicationName",
                        "ApplicationVersion",
                        "Visible",
                        "Usable",
                        "ProcessingPass",
                        "ExtraPackages",
                        "DDDB",
                        "CONDDB",
                        "DQTag",
                    ],
                    "Records": [
                        [
                            998,
                            "",
                            "",
                            "x86_64-slc6-gcc49-opt",
                            "",
                            "N",
                            "Stripping28",
                            "DaVinci",
                            "v41r3",
                            "Yes",
                            "Yes",
                            "Stripping28",
                            "AppConfig.v3r306",
                            "dddb-20150724",
                            "cond-20161011",
                            "",
                        ]
                    ],
                },
            }
        elif stepID == {"StepId": 997}:
            return {
                "OK": True,
                "Value": {
                    "TotalRecords": 1,
                    "ParameterNames": [
                        "StepId",
                        "OptionFiles",
                        "ExtraOptions",
                        "SystemConfig",
                        "mcTCK",
                        "isMulticore",
                        "StepName",
                        "ApplicationName",
                        "ApplicationVersion",
                        "Visible",
                        "Usable",
                        "ProcessingPass",
                        "ExtraPackages",
                        "DDDB",
                        "CONDDB",
                        "DQTag",
                    ],
                    "Records": [
                        [
                            997,
                            "$APPCONFIGOPTS/Brunel/DataType-2016.py;$APPCONFIGOPTS/Brunel/PbPb-GECs.py",
                            "",
                            "x86_64-slc6-gcc49-opt",
                            "",
                            "N",
                            "Stripping28",
                            "DaVinci",
                            "v41r3",
                            "Yes",
                            "Yes",
                            "Stripping28",
                            "AppConfig.v3r306",
                            "dddb-20150724",
                            "cond-20161011",
                            "",
                        ]
                    ],
                },
            }
        elif stepID == {"StepId": 996}:
            options996 = "$APPCONFIGOPTS/Brunel/DataType-2016.py;"
            options996 += "$APPCONFIGOPTS/Persistency/Compression-ZLIB-1.py;"
            options996 += "$APPCONFIGOPTS/Brunel/PbPb-GECs.py"
            return {
                "OK": True,
                "Value": {
                    "TotalRecords": 1,
                    "ParameterNames": [
                        "StepId",
                        "OptionFiles",
                        "ExtraOptions",
                        "SystemConfig",
                        "mcTCK",
                        "isMulticore",
                        "StepName",
                        "ApplicationName",
                        "ApplicationVersion",
                        "Visible",
                        "Usable",
                        "ProcessingPass",
                        "ExtraPackages",
                        "DDDB",
                        "CONDDB",
                        "DQTag",
                    ],
                    "Records": [
                        [
                            996,
                            options996,
                            "",
                            "x86_64-slc6-gcc49-opt",
                            "",
                            "N",
                            "Stripping28",
                            "DaVinci",
                            "v41r3",
                            "Yes",
                            "Yes",
                            "Stripping28",
                            "AppConfig.v3r306",
                            "dddb-20150724",
                            "cond-20161011",
                            "",
                        ]
                    ],
                },
            }
        elif stepID == {"StepId": 1098}:
            options1098 = "$APPCONFIGOPTS/Brunel/DataType-2016.py;"
            options1098 += "$APPCONFIGOPTS/Persistency/Compression-ZLIB-1.py;"
            options1098 += "$APPCONFIGOPTS/Persistency/Compression-LZMA-4.py"
            return {
                "OK": True,
                "Value": {
                    "TotalRecords": 1,
                    "ParameterNames": [
                        "StepId",
                        "OptionFiles",
                        "ExtraOptions",
                        "SystemConfig",
                        "mcTCK",
                        "isMulticore",
                        "StepName",
                        "ApplicationName",
                        "ApplicationVersion",
                        "Visible",
                        "Usable",
                        "ProcessingPass",
                        "ExtraPackages",
                        "DDDB",
                        "CONDDB",
                        "DQTag",
                    ],
                    "Records": [
                        [
                            1098,
                            options1098,
                            "",
                            "x86_64-slc6-gcc49-opt",
                            "",
                            "N",
                            "Stripping28",
                            "DaVinci",
                            "v41r3",
                            "Yes",
                            "Yes",
                            "Stripping28",
                            "AppConfig.v3r306",
                            "dddb-20150724",
                            "cond-20161011",
                            "",
                        ]
                    ],
                },
            }
        elif stepID == {"StepId": 1099}:
            options1099 = "$APPCONFIGOPTS/Brunel/DataType-2016.py;"
            options1099 += "$APPCONFIGOPTS/Persistency/Compression-ZLIB-1.py;"
            options1099 += "$APPCONFIGOPTS/Persistency/Compression-LZMA-4.py"
            return {
                "OK": True,
                "Value": {
                    "TotalRecords": 1,
                    "ParameterNames": [
                        "StepId",
                        "OptionFiles",
                        "ExtraOptions",
                        "SystemConfig",
                        "mcTCK",
                        "isMulticore",
                        "StepName",
                        "ApplicationName",
                        "ApplicationVersion",
                        "Visible",
                        "Usable",
                        "ProcessingPass",
                        "ExtraPackages",
                        "DDDB",
                        "CONDDB",
                        "DQTag",
                    ],
                    "Records": [
                        [
                            1099,
                            options1099,
                            "",
                            "x86_64-slc6-gcc49-opt",
                            "",
                            "N",
                            "Stripping28",
                            "DaVinci",
                            "v41r3",
                            "Yes",
                            "Yes",
                            "Stripping28",
                            "AppConfig.v3r306",
                            "dddb-20150724",
                            "cond-20161011",
                            "",
                        ]
                    ],
                },
            }

    def getStepInputFiles(self, stepID):
        if stepID == 123:
            return {
                "OK": True,
                "Value": {"TotalRecords": 7, "ParameterNames": ["FileType", "Visible"], "Records": [["SDST", "Y"]]},
            }

        if stepID == 456:
            return {
                "OK": True,
                "Value": {
                    "TotalRecords": 7,
                    "ParameterNames": ["FileType", "Visible"],
                    "Records": [["BHADRON.DST", "Y"], ["CALIBRATION.DST", "Y"]],
                },
            }
        if stepID == 789:
            return {
                "OK": True,
                "Value": {
                    "TotalRecords": 7,
                    "ParameterNames": ["FileType", "Visible"],
                    "Records": [["DAVINCIHIST", "Y"], ["BRUNELHIST", "Y"]],
                },
            }
        if stepID == 125080:
            return {
                "OK": True,
                "Value": {"TotalRecords": 7, "ParameterNames": ["FileType", "Visible"], "Records": [["", "Y"]]},
            }
        if stepID == 124620 or stepID == 124621:
            return {
                "OK": True,
                "Value": {"TotalRecords": 7, "ParameterNames": ["FileType", "Visible"], "Records": [["SIM", "N"]]},
            }
        if stepID == 999 or stepID == 998 or stepID == 997 or stepID == 996 or stepID == 1098 or stepID == 1099:
            return {
                "OK": True,
                "Value": {"TotalRecords": 7, "ParameterNames": ["FileType", "Visible"], "Records": [["SDST", "Y"]]},
            }

    def getStepOutputFiles(self, stepID):
        if stepID == 123:
            return {
                "OK": True,
                "Value": {
                    "TotalRecords": 7,
                    "ParameterNames": ["FileType", "Visible"],
                    "Records": [["BHADRON.DST", "Y"], ["CALIBRATION.DST", "Y"]],
                },
            }
        if stepID == 456:
            return {
                "OK": True,
                "Value": {
                    "TotalRecords": 7,
                    "ParameterNames": ["FileType", "Visible"],
                    "Records": [["BHADRON.DST", "Y"], ["CALIBRATION.DST", "Y"]],
                },
            }
        if stepID == 789:
            return {
                "OK": True,
                "Value": {"TotalRecords": 7, "ParameterNames": ["FileType", "Visible"], "Records": [["ROOT", "Y"]]},
            }
        if stepID == 125080:
            return {
                "OK": True,
                "Value": {"TotalRecords": 7, "ParameterNames": ["FileType", "Visible"], "Records": [["SIM", "Y"]]},
            }
        if stepID == 124620:
            return {
                "OK": True,
                "Value": {"TotalRecords": 7, "ParameterNames": ["FileType", "Visible"], "Records": [["DIGI", "N"]]},
            }
        if stepID == 124621:
            return {
                "OK": True,
                "Value": {"TotalRecords": 7, "ParameterNames": ["FileType", "Visible"], "Records": [["MDF", "Y"]]},
            }
        if stepID == 999 or stepID == 998 or stepID == 997 or stepID == 996 or stepID == 1098 or stepID == 1099:
            return {
                "OK": True,
                "Value": {
                    "TotalRecords": 7,
                    "ParameterNames": ["FileType", "Visible"],
                    "Records": [["BHADRON.DST", "Y"], ["CALIBRATION.DST", "Y"]],
                },
            }


# Some steps definitions, for testing purposes

options125080 = "$APPCONFIGOPTS/Gauss/Sim08-Beam4000GeV-mu100-2012-nu2.5.py;"
options125080 += "$DECFILESROOT/options/11102400.py;"
options125080 += "$LBPYTHIA8ROOT/options/Pythia8.py;"
options125080 += "$APPCONFIGOPTS/Gauss/G4PL_FTFP_BERT_EmNoCuts.py;"
step125080 = {
    "StepId": 125080,
    "StepName": "Sim08a",
    "ApplicationName": "Gauss",
    "ApplicationVersion": "v45r3",
    "ExtraOptions": "",
    "EventTimeout": "",
    "OptionFiles": options125080,
    "Usable": "Yes",
    "Visible": "Y",
    "ExtraPackages": "AppConfig.v3r171",
    "ProcessingPass": "Sim08a",
    "OptionsFormat": "",
    "prodStepID": "125080['']",
    "SystemConfig": "x86_64-slc5-gcc43-opt",
    "DDDB": "Sim08-20130503-1",
    "CONDDB": "Sim08-20130503-1-vc-mu100",
    "DQTag": "",
    "isMulticore": "N",
    "mcTCK": "",
    "visibilityFlag": [{"Visible": "N", "FileType": "SIM"}],
    "fileTypesIn": [""],
    "fileTypesOut": ["SIM"],
}

stepMC = {
    "StepId": 123,
    "StepName": "Stripping14-Stripping",
    "ApplicationName": "Gauss",
    "ApplicationVersion": "v2r2",
    "ExtraOptions": "",
    "EventTimeout": "",
    "OptionFiles": "optsFiles",
    "Visible": "Yes",
    "ExtraPackages": "eps",
    "ProcessingPass": "procPass",
    "OptionsFormat": "",
    "prodStepID": "123['SDST']",
    "SystemConfig": "",
    "DDDB": "",
    "CONDDB": "123456",
    "DQTag": "",
    "isMulticore": "N",
    "mcTCK": "",
    "visibilityFlag": [{"Visible": "N", "FileType": "SIM"}],
    "fileTypesIn": [],
    "fileTypesOut": ["SIM"],
}

stepMC2 = {
    "StepId": 123,
    "StepName": "Stripping14-Stripping",
    "ApplicationName": "DaVinci",
    "ApplicationVersion": "v2r2",
    "ExtraOptions": "",
    "EventTimeout": "",
    "OptionFiles": "optsFiles",
    "Visible": "Yes",
    "ExtraPackages": "eps",
    "ProcessingPass": "procPass",
    "OptionsFormat": "",
    "prodStepID": "123['SDST']",
    "SystemConfig": "",
    "DDDB": "",
    "CONDDB": "123456",
    "DQTag": "",
    "isMulticore": "N",
    "mcTCK": "",
    "visibilityFlag": [{"Visible": "Y", "FileType": "DST"}],
    "fileTypesIn": ["SIM"],
    "fileTypesOut": ["DST"],
}

step1Dict = {
    "StepId": 123,
    "StepName": "Stripping14-Stripping",
    "ApplicationName": "DaVinci",
    "ApplicationVersion": "v2r2",
    "ExtraOptions": "",
    "EventTimeout": "",
    "OptionFiles": "optsFiles",
    "Visible": "Yes",
    "ExtraPackages": "eps",
    "ProcessingPass": "procPass",
    "OptionsFormat": "",
    "prodStepID": "123['SDST']",
    "SystemConfig": "",
    "DDDB": "",
    "CONDDB": "123456",
    "DQTag": "",
    "isMulticore": "N",
    "mcTCK": "",
    "visibilityFlag": [{"Visible": "N", "FileType": "BHADRON.DST"}, {"Visible": "N", "FileType": "CALIBRATION.DST"}],
    "fileTypesIn": ["SDST"],
    "fileTypesOut": ["BHADRON.DST", "CALIBRATION.DST"],
}

step2Dict = {
    "StepId": 456,
    "StepName": "Merge",
    "ApplicationName": "LHCb",
    "ApplicationVersion": "v1r2",
    "ExtraOptions": "",
    "EventTimeout": "",
    "OptionFiles": "optsFiles",
    "Visible": "Yes",
    "ExtraPackages": "eps",
    "ProcessingPass": "procPass",
    "OptionsFormat": "",
    "SystemConfig": "x86_64-slc6-gcc49-opt",
    "prodStepID": "456['BHADRON.DST', 'CALIBRATION.DST']",
    "DDDB": "",
    "CONDDB": "123456",
    "DQTag": "",
    "isMulticore": "N",
    "mcTCK": "",
    "visibilityFlag": [{"Visible": "N", "FileType": "BHADRON.DST"}, {"Visible": "N", "FileType": "CALIBRATION.DST"}],
    "fileTypesIn": ["BHADRON.DST", "CALIBRATION.DST"],
    "fileTypesOut": ["BHADRON.DST", "CALIBRATION.DST"],
}

step124620Dict = {
    "StepId": 124620,
    "StepName": "Digi13",
    "ApplicationName": "Boole",
    "ApplicationVersion": "v26r3",
    "ExtraOptions": "",
    "EventTimeout": "",
    "OptionFiles": "optsFiles",
    "Visible": "N",
    "ExtraPackages": "AppConfig.v3r164",
    "ProcessingPass": "Digi13",
    "OptionsFormat": "",
    "prodStepID": "124620['SIM']",
    "SystemConfig": "x86_64-slc5-gcc43-opt",
    "DDDB": "Sim08-20130503-1",
    "CONDDB": "Sim08-20130503-1-vc-mu100",
    "DQTag": "",
    "isMulticore": "N",
    "mcTCK": "",
    "Usable": "Yes",
    "visibilityFlag": [{"Visible": "N", "FileType": "DIGI"}],
    "fileTypesIn": ["SIM"],
    "fileTypesOut": ["DIGI"],
}

step124621Dict = {
    "StepId": 124621,
    "StepName": "Digi13",
    "ApplicationName": "Brunel",
    "ApplicationVersion": "v26r3",
    "ExtraOptions": "",
    "EventTimeout": "",
    "OptionFiles": "optsFiles",
    "Visible": "N",
    "ExtraPackages": "AppConfig.v3r164",
    "ProcessingPass": "Digi13",
    "OptionsFormat": "",
    "prodStepID": "124621['SIM']",
    "SystemConfig": "x86_64-slc5-gcc43-opt",
    "DDDB": "Sim08-20130503-1",
    "CONDDB": "Sim08-20130503-1-vc-mu100",
    "DQTag": "",
    "isMulticore": "N",
    "mcTCK": "",
    "Usable": "Yes",
    "visibilityFlag": [{"Visible": "Y", "FileType": "MDF"}],
    "fileTypesIn": ["SIM"],
    "fileTypesOut": ["MDF"],
}

stepStripp = {
    "ApplicationName": "DaVinci",
    "Usable": "Yes",
    "StepId": 123,
    "ApplicationVersion": "v28r3p1",
    "ExtraPackages": "AppConfig.v3r104",
    "StepName": "Stripping14-Merging",
    "SystemConfig": "",
    "ProcessingPass": "Merging",
    "Visible": "N",
    "DDDB": "head-20110302",
    "mcTCK": "",
    "OptionFiles": "$APPCONFIGOPTS/Merging/DV-Stripping14-Merging.py",
    "CONDDB": "head-20110407",
    "visibilityFlag": [],
    "fileTypesIn": ["SDST"],
    "fileTypesOut": ["BHADRON.DST", "CALIBRATION.DST", "CHARM.MDST", "CHARMCOMPLETEEVENT.DST"],
}
mergeStep = {
    "ApplicationName": "DaVinci",
    "Usable": "Yes",
    "StepId": 456,
    "ApplicationVersion": "v28r3p1",
    "ExtraPackages": "AppConfig.v3r104",
    "StepName": "Stripping14-Merging",
    "SystemConfig": "",
    "ProcessingPass": "Merging",
    "Visible": "N",
    "DDDB": "head-20110302",
    "mcTCK": "",
    "OptionFiles": "$APPCONFIGOPTS/Merging/DV-Stripping14-Merging.py",
    "CONDDB": "head-20110407",
    "visibilityFlag": [],
    "fileTypesIn": ["BHADRON.DST", "CALIBRATION.DST", "PID.MDST"],
    "fileTypesOut": ["BHADRON.DST", "CALIBRATION.DST", "PID.MDST"],
}
mergeStepBHADRON = {
    "ApplicationName": "DaVinci",
    "Usable": "Yes",
    "StepId": 456,
    "ApplicationVersion": "v28r3p1",
    "ExtraPackages": "AppConfig.v3r104",
    "StepName": "Stripping14-Merging",
    "SystemConfig": "",
    "prodStepID": "456['BHADRON.DST']",
    "ProcessingPass": "Merging",
    "Visible": "N",
    "DDDB": "head-20110302",
    "mcTCK": "",
    "OptionFiles": "$APPCONFIGOPTS/Merging/DV-Stripping14-Merging.py",
    "CONDDB": "head-20110407",
    "visibilityFlag": [],
    "fileTypesIn": ["BHADRON.DST"],
    "fileTypesOut": ["BHADRON.DST"],
}
mergeStepCALIBRA = {
    "ApplicationName": "DaVinci",
    "Usable": "Yes",
    "StepId": 456,
    "ApplicationVersion": "v28r3p1",
    "ExtraPackages": "AppConfig.v3r104",
    "StepName": "Stripping14-Merging",
    "SystemConfig": "",
    "prodStepID": "456['CALIBRATION.DST']",
    "ProcessingPass": "Merging",
    "Visible": "N",
    "DDDB": "head-20110302",
    "mcTCK": "",
    "OptionFiles": "$APPCONFIGOPTS/Merging/DV-Stripping14-Merging.py",
    "CONDDB": "head-20110407",
    "visibilityFlag": [],
    "fileTypesIn": ["CALIBRATION.DST"],
    "fileTypesOut": ["CALIBRATION.DST"],
}
mergeStepPIDMDST = {
    "ApplicationName": "DaVinci",
    "Usable": "Yes",
    "StepId": 456,
    "ApplicationVersion": "v28r3p1",
    "ExtraPackages": "AppConfig.v3r104",
    "StepName": "Stripping14-Merging",
    "SystemConfig": "",
    "prodStepID": "456['PID.MDST']",
    "ProcessingPass": "Merging",
    "Visible": "N",
    "DDDB": "head-20110302",
    "mcTCK": "",
    "OptionFiles": "$APPCONFIGOPTS/Merging/DV-Stripping14-Merging.py",
    "CONDDB": "head-20110407",
    "visibilityFlag": [],
    "fileTypesIn": ["PID.MDST"],
    "fileTypesOut": ["PID.MDST"],
}

stepHistoMergingDict = {
    "StepId": 789,
    "StepName": "MergeHisto",
    "ApplicationName": "Noether",
    "ApplicationVersion": "v1r2",
    "ExtraOptions": "",
    "EventTimeout": "",
    "OptionFiles": "optsFiles",
    "Visible": "Yes",
    "ExtraPackages": "eps",
    "ProcessingPass": "procPass",
    "OptionsFormat": "",
    "SystemConfig": "x86_64-slc6-gcc49-opt",
    "prodStepID": "789['DAVINCIHIST', 'BRUNELHIST']",
    "DDDB": "",
    "CONDDB": "123456",
    "DQTag": "",
    "isMulticore": "N",
    "mcTCK": "",
    "visibilityFlag": [{"Visible": "Y", "FileType": "ROOT"}],
    "fileTypesIn": ["DAVINCIHIST", "BRUNELHIST"],
    "fileTypesOut": ["ROOT"],
}



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
:mod: UserStorageQuotaAgent

.. module: UserStorageQuotaAgent

:synopsis: UserStorageQuotaAgent obtains the usage by each user from the StorageUsageDB
  and compares with a quota present in the CS.
"""
from DIRAC import gConfig, S_OK
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.FrameworkSystem.Client.NotificationClient import NotificationClient
from LHCbDIRAC.DataManagementSystem.DB.StorageUsageDB import StorageUsageDB
from LHCbDIRAC.DataManagementSystem.Client.StorageUsageClient import StorageUsageClient

AGENT_NAME = "DataManagement/UserStorageQuotaAgent"


class UserStorageQuotaAgent(AgentModule):
    """.. class:: UserStorageQuotaAgent.

    :param int deafultQuota: default quota in MB
    :param NotificationClient notificationClient: NotificationClient instance
    :param LHCbDIRAC.DataManagementSystem.DB.StorageUsageDB.StorageUsageDB storageUsageDB:
           StorageUsageDB or RPC client pointing to StorageUsageDB
    """

    defaultQuota = 1000
    notificationClient = None
    storageUsageDB = None

    def __init__(self, *args, **kwargs):
        """c'tor."""
        AgentModule.__init__(self, *args, **kwargs)

        self.notificationClient = NotificationClient()
        try:
            self.storageUsageDB = StorageUsageDB()
        except SystemExit:
            self.storageUsageDB = StorageUsageClient()

        self.defaultQuota = gConfig.getValue("/Registry/DefaultStorageQuota", self.defaultQuota)  # Default is 1TB

    def initialize(self):
        """agent initialisation.

        :param self: self reference
        """
        # This sets the Default Proxy to used as that defined under
        # /Operations/Shifter/DataManager
        # the shifterProxy option in the Configuration can be used to change this default.
        self.am_setOption("shifterProxy", "DataManager")

        self.log.info("initialize: Default quota found to be %d GB" % self.defaultQuota)
        return S_OK()

    def execute(self):
        """execution of one cycle.

        :param self: self reference
        """
        res = self.storageUsageDB.getUserStorageUsage()
        usageDict = res["Value"]

        byteToGB = 1000 * 1000 * 1000.0

        managerMsg = ""
        errorMsg = ""
        self.log.info(f"Determining quota usage for {len(usageDict)} users.")
        for userName in sorted(usageDict):
            usageGB = usageDict[userName] / byteToGB
            res = gConfig.getOptionsDict(f"/Registry/Users/{userName}")
            if not res["OK"]:
                msg = f"Username not found in the CS: {userName} using {usageGB:.2f} GB"
                errorMsg += msg + "\n"
                self.log.error(msg)
                continue
            elif "Email" not in res["Value"]:
                msg = f"CS does not contain email information for user {userName}"
                errorMsg += msg + "\n"
                self.log.error(msg)
                continue
            elif "Quota" not in res["Value"]:
                userQuota = float(self.defaultQuota)
            else:
                userQuota = float(res["Value"]["Quota"])
            userMail = res["Value"]["Email"]
            # Different behaviour for 90% exceeded, 110% exceeded and 150% exceeded
            msg = None
            if (1.5 * userQuota) < usageGB:
                msg = "%s is at %d%s of quota %d GB (%.1f GB)." % (
                    userName,
                    (usageGB * 100) / userQuota,
                    "%",
                    userQuota,
                    usageGB,
                )
                self.log.info(msg)
                self.sendBlockedMail(userName, userMail, userQuota, usageGB)
                self.log.info("!!!!!!!!!!!!!!!!!!!!!!!!REMEMBER TO MODIFY THE ACLs and STATUS HERE!!!!!!!!!!!!!!!!!")
            elif (1.0 * userQuota) < usageGB:
                msg = "%s is at %d%s of quota %d GB (%.1f GB)." % (
                    userName,
                    (usageGB * 100) / userQuota,
                    "%",
                    userQuota,
                    usageGB,
                )
                self.log.info(msg)
                self.sendSecondWarningMail(userName, userMail, userQuota, usageGB)
            elif (0.9 * userQuota) < usageGB:
                msg = "%s is at %d%s of quota %d GB (%.1f GB)." % (
                    userName,
                    (usageGB * 100) / userQuota,
                    "%",
                    userQuota,
                    usageGB,
                )
                self.log.info(msg)
                self.sendFirstWarningMail(userName, userMail, userQuota, usageGB)
            if msg:
                managerMsg += msg + "\n"
        if managerMsg or errorMsg:
            if managerMsg:
                managerMsg = (
                    "Mails have been sent to the following list of users "
                    + "being close to or above quota:\n\n"
                    + managerMsg
                )
            if errorMsg:
                managerMsg += "\nThe following errors have been found by the UserStorageQuotaAgent:\n" + errorMsg
            fromAddress = "LHCb Data Manager <lhcb-datamanagement@cern.ch>"
            toAddress = "lhcb-datamanagement@cern.ch"
            self.notificationClient.sendMail(toAddress, "User quota warnings", managerMsg, fromAddress)
        return S_OK()

    def sendFirstWarningMail(self, userName, userMail, quota, usage):
        """first warning email.

        :param self: self reference
        :param str userName: DIRAC user name
        :param str userMail: email address
        :param int quota: default quota
        :param float usage: space currently used
        """
        msgbody = """
This mail has been generated automatically.

You have received this mail because you are approaching your Grid storage usage quota of {} GB.

You are currently using {:.1f} GB.

Please reduce you usage by removing some files. If you have reduced your usage in the last 24 hours
please ignore this message.

Explanations can be found at https://twiki.cern.ch/twiki/bin/view/LHCb/GridStorageQuota
""".format(
            int(quota),
            usage,
        )
        fromAddress = "LHCb Data Manager <lhcb-datamanagement@cern.ch>"
        subject = f"Grid storage use near quota ({userName})"
        toAddress = userMail
        self.notificationClient.sendMail(toAddress, subject, msgbody, fromAddress)

    def sendSecondWarningMail(self, userName, userMail, quota, usage):
        """second warning email.

        :param self: self reference
        :param str userName: DIRAC user name
        :param str userMail: email address
        :param int quota: default quota
        :param float usage: space currently used
        """
        msgbody = """
This mail has been generated automatically.

You have received this mail because your Grid storage usage has exceeded your quota of {}GB.

You are currently using {:.1f} GB.

Please reduce you usage by removing some files. If you have reduced your usage in the last 24 hours
please ignore this message.

Explanations can be found at https://twiki.cern.ch/twiki/bin/view/LHCb/GridStorageQuota
""".format(
            int(quota),
            usage,
        )
        fromAddress = "LHCb Data Manager <lhcb-datamanagement@cern.ch>"
        subject = f"Grid storage use over quota ({userName})"
        toAddress = userMail
        self.notificationClient.sendMail(toAddress, subject, msgbody, fromAddress)

    def sendBlockedMail(self, userName, userMail, quota, usage):
        """send blocked email.

        :param self: self reference
        :param str userName: DIRAC user name
        :param str userMail: email adress
        :param int quota: default quota
        :param float usage: space used
        """
        msgbody = """
This mail has been generated automatically.

You have received this mail because your Grid storage usage has exceeded your quota of {} GB.

You are currently using {:.1f} GB.

Your account could soon been given a lower priority and your jobs will run at a lower pace if you
don't create space.
If you have reduced your usage in the last 24 hours please ignore this message.
Explanations can be found at https://twiki.cern.ch/twiki/bin/view/LHCb/GridStorageQuota
""".format(
            int(quota),
            usage,
        )

        fromAddress = "LHCb Data Manager <lhcb-datamanagement@cern.ch>"
        subject = f"Grid storage use blocked ({userName})"
        toAddress = userMail
        self.notificationClient.sendMail(toAddress, subject, msgbody, fromAddress)



###############################################################################
# (c) Copyright 2023 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import pytest
from DIRAC import S_OK

import LHCbDIRAC
from LHCbDIRAC.BookkeepingSystem.DB.OracleBookkeepingDB import BookkeepingResultMismatch, ProxyMethod
from LHCbDIRAC.BookkeepingSystem.DB.LegacyOracleBookkeepingDB import LegacyOracleBookkeepingDB
from LHCbDIRAC.BookkeepingSystem.DB.NewOracleBookkeepingDB import NewOracleBookkeepingDB


class DemoException(Exception):
    pass


def _raising_function(self):
    raise DemoException("This should never be reached")


@pytest.fixture
def fake_bkdb(monkeypatch, bkdb):
    """Return an instance of OracleBookkeepingDB with fake methods monkeypatched."""
    monkeypatch.setattr(
        LegacyOracleBookkeepingDB, "fakeDuplicated", lambda _: S_OK("resultLegacyDuplicated"), raising=False
    )
    monkeypatch.setattr(NewOracleBookkeepingDB, "fakeDuplicated", lambda _: S_OK("resultNewDuplicated"), raising=False)

    monkeypatch.setattr(LegacyOracleBookkeepingDB, "fakeMatching", lambda _: S_OK("resultIdentical"), raising=False)
    monkeypatch.setattr(NewOracleBookkeepingDB, "fakeMatching", lambda _: S_OK("resultIdentical"), raising=False)

    monkeypatch.setattr(
        LegacyOracleBookkeepingDB, "fakeLegacyOnly", lambda _: S_OK("resultLegacyLegacyOnly"), raising=False
    )

    monkeypatch.setattr(NewOracleBookkeepingDB, "fakeNewOnly", lambda _: S_OK("resultNewNewOnly"), raising=False)

    monkeypatch.setattr(
        LegacyOracleBookkeepingDB, "fakeNewRaises", lambda _: S_OK("resultLegacyNewRaises"), raising=False
    )
    monkeypatch.setattr(NewOracleBookkeepingDB, "fakeNewRaises", _raising_function, raising=False)

    monkeypatch.setattr(LegacyOracleBookkeepingDB, "fakeLegacyRaises", _raising_function, raising=False)
    monkeypatch.setattr(
        NewOracleBookkeepingDB, "fakeLegacyRaises", lambda _: S_OK("resultNewLegacyRaises"), raising=False
    )

    monkeypatch.setattr(
        LHCbDIRAC.BookkeepingSystem.DB.OracleBookkeepingDB,
        "DUPLICATION_SAFE_METHODS",
        ["fakeDuplicated", "fakeLegacyOnly", "fakeNewOnly", "fakeMatching"],
    )

    yield bkdb


def test_preferred_implementation(monkeypatch, fake_bkdb):
    assert not isinstance(fake_bkdb.fakeDuplicated, ProxyMethod)
    assert fake_bkdb.fakeDuplicated() == S_OK("resultLegacyDuplicated")

    monkeypatch.setattr(LHCbDIRAC.BookkeepingSystem.DB.OracleBookkeepingDB, "DO_COMPARE", True)

    assert isinstance(fake_bkdb.fakeDuplicated, ProxyMethod)
    assert fake_bkdb.fakeDuplicated() == S_OK("resultLegacyDuplicated")

    monkeypatch.setattr(LHCbDIRAC.BookkeepingSystem.DB.OracleBookkeepingDB, "METHODS_PREFER_NEW", ["fakeDuplicated"])

    assert isinstance(fake_bkdb.fakeDuplicated, ProxyMethod)
    assert fake_bkdb.fakeDuplicated() == S_OK("resultNewDuplicated")

    monkeypatch.setattr(LHCbDIRAC.BookkeepingSystem.DB.OracleBookkeepingDB, "DO_COMPARE", False)

    assert not isinstance(fake_bkdb.fakeDuplicated, ProxyMethod)
    assert fake_bkdb.fakeDuplicated() == S_OK("resultNewDuplicated")


def test_exception_not_raised(monkeypatch, fake_bkdb):
    monkeypatch.setattr(LHCbDIRAC.BookkeepingSystem.DB.OracleBookkeepingDB, "FAIL_ON_DIFFERENCE", True)

    assert not isinstance(fake_bkdb.fakeMatching, ProxyMethod)
    assert fake_bkdb.fakeMatching() == S_OK("resultIdentical")

    monkeypatch.setattr(LHCbDIRAC.BookkeepingSystem.DB.OracleBookkeepingDB, "DO_COMPARE", True)

    assert isinstance(fake_bkdb.fakeMatching, ProxyMethod)
    assert fake_bkdb.fakeMatching() == S_OK("resultIdentical")

    monkeypatch.setattr(LHCbDIRAC.BookkeepingSystem.DB.OracleBookkeepingDB, "METHODS_PREFER_NEW", ["fakeMatching"])

    assert isinstance(fake_bkdb.fakeMatching, ProxyMethod)
    assert fake_bkdb.fakeMatching() == S_OK("resultIdentical")

    monkeypatch.setattr(LHCbDIRAC.BookkeepingSystem.DB.OracleBookkeepingDB, "DO_COMPARE", False)

    assert not isinstance(fake_bkdb.fakeMatching, ProxyMethod)
    assert fake_bkdb.fakeMatching() == S_OK("resultIdentical")


def test_exception_raised(monkeypatch, fake_bkdb):
    monkeypatch.setattr(LHCbDIRAC.BookkeepingSystem.DB.OracleBookkeepingDB, "FAIL_ON_DIFFERENCE", True)

    assert not isinstance(fake_bkdb.fakeDuplicated, ProxyMethod)
    assert fake_bkdb.fakeDuplicated() == S_OK("resultLegacyDuplicated")

    monkeypatch.setattr(LHCbDIRAC.BookkeepingSystem.DB.OracleBookkeepingDB, "DO_COMPARE", True)

    assert isinstance(fake_bkdb.fakeDuplicated, ProxyMethod)
    with pytest.raises(BookkeepingResultMismatch):
        fake_bkdb.fakeDuplicated()

    monkeypatch.setattr(LHCbDIRAC.BookkeepingSystem.DB.OracleBookkeepingDB, "METHODS_PREFER_NEW", ["fakeDuplicated"])

    assert isinstance(fake_bkdb.fakeDuplicated, ProxyMethod)
    with pytest.raises(BookkeepingResultMismatch):
        fake_bkdb.fakeDuplicated()

    monkeypatch.setattr(LHCbDIRAC.BookkeepingSystem.DB.OracleBookkeepingDB, "DO_COMPARE", False)

    assert not isinstance(fake_bkdb.fakeDuplicated, ProxyMethod)
    assert fake_bkdb.fakeDuplicated() == S_OK("resultNewDuplicated")


@pytest.mark.parametrize(
    "name,expected_result",
    [
        ["fakeLegacyOnly", "resultLegacyLegacyOnly"],
        ["fakeNewOnly", "resultNewNewOnly"],
    ],
)
def test_one_only(monkeypatch, fake_bkdb, name, expected_result):
    assert not isinstance(getattr(fake_bkdb, name), ProxyMethod)
    assert getattr(fake_bkdb, name)() == S_OK(expected_result)

    monkeypatch.setattr(LHCbDIRAC.BookkeepingSystem.DB.OracleBookkeepingDB, "DO_COMPARE", True)

    assert not isinstance(getattr(fake_bkdb, name), ProxyMethod)
    assert getattr(fake_bkdb, name)() == S_OK(expected_result)

    monkeypatch.setattr(LHCbDIRAC.BookkeepingSystem.DB.OracleBookkeepingDB, "METHODS_PREFER_NEW", ["fakeLegacyOnly"])

    assert not isinstance(getattr(fake_bkdb, name), ProxyMethod)
    assert getattr(fake_bkdb, name)() == S_OK(expected_result)


def test_new_raising(monkeypatch, fake_bkdb):
    assert not isinstance(fake_bkdb.fakeNewRaises, ProxyMethod)
    assert fake_bkdb.fakeNewRaises() == S_OK("resultLegacyNewRaises")

    monkeypatch.setattr(LHCbDIRAC.BookkeepingSystem.DB.OracleBookkeepingDB, "DO_COMPARE", True)
    assert fake_bkdb.fakeNewRaises() == S_OK("resultLegacyNewRaises")

    monkeypatch.setattr(
        LHCbDIRAC.BookkeepingSystem.DB.OracleBookkeepingDB, "DUPLICATION_SAFE_METHODS", ["fakeNewRaises"]
    )
    with pytest.raises(DemoException):
        fake_bkdb.fakeNewRaises()

    monkeypatch.setattr(LHCbDIRAC.BookkeepingSystem.DB.OracleBookkeepingDB, "DO_COMPARE", False)
    assert fake_bkdb.fakeNewRaises() == S_OK("resultLegacyNewRaises")

    monkeypatch.setattr(LHCbDIRAC.BookkeepingSystem.DB.OracleBookkeepingDB, "METHODS_PREFER_NEW", ["fakeNewRaises"])
    with pytest.raises(DemoException):
        fake_bkdb.fakeNewRaises()


def test_legacy_raising(monkeypatch, fake_bkdb):
    assert not isinstance(fake_bkdb.fakeNewRaises, ProxyMethod)
    with pytest.raises(DemoException):
        fake_bkdb.fakeLegacyRaises()

    monkeypatch.setattr(LHCbDIRAC.BookkeepingSystem.DB.OracleBookkeepingDB, "DO_COMPARE", True)
    with pytest.raises(DemoException):
        fake_bkdb.fakeLegacyRaises()

    monkeypatch.setattr(
        LHCbDIRAC.BookkeepingSystem.DB.OracleBookkeepingDB, "DUPLICATION_SAFE_METHODS", ["fakeLegacyRaises"]
    )
    with pytest.raises(DemoException):
        fake_bkdb.fakeLegacyRaises()

    monkeypatch.setattr(LHCbDIRAC.BookkeepingSystem.DB.OracleBookkeepingDB, "DO_COMPARE", False)
    with pytest.raises(DemoException):
        fake_bkdb.fakeLegacyRaises()

    monkeypatch.setattr(LHCbDIRAC.BookkeepingSystem.DB.OracleBookkeepingDB, "METHODS_PREFER_NEW", ["fakeLegacyRaises"])
    assert fake_bkdb.fakeLegacyRaises() == S_OK("resultNewLegacyRaises")



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
Returns metadata of the job step(s) that created a (list of) LFNs,
or all steps for a (list of) JobID
"""

from DIRAC.Core.Base.Script import Script
from DIRAC.Core.Utilities.ReturnValues import returnValueOrRaise


@Script()
def main():
    from DIRAC import gLogger
    from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript, printDMResult

    bkScript = DMScript()
    bkScript.registerFileSwitches()
    Script.registerSwitch("", "Summary", "   Only report job IDs creating the LFN(s)")
    Script.setUsageMessage(
        __doc__
        + "\n".join(
            [
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile] ... [LFN|File|JobID]",
                "Arguments:",
                "  LFN:      Logical File Name",
                "  File:     Name of the file with a list of LFNs",
                "  JobID:    DIRAC job identifier",
            ]
        )
    )

    Script.parseCommandLine(ignoreErrors=True)
    args = Script.getPositionalArgs()
    jobIDList = []
    for arg in args:
        if arg.isdigit():
            jobIDList.append(int(arg))
        else:
            bkScript.setLFNsFromFile(arg)
    lfnList = bkScript.getOption("LFNs", [])
    if not lfnList and not jobIDList:
        Script.showHelp(exitCode=1)
    summary = False
    for switch in Script.getUnprocessedSwitches():
        if switch[0] == "Summary":
            summary = True

    from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

    if lfnList:
        retVal = BookkeepingClient().bulkJobInfo(lfnList)
    elif jobIDList:
        retVal = BookkeepingClient().bulkJobInfo({"jobId": jobIDList})
        summary = False
    else:
        raise Exception("No LFNs or JobIDs given")
    jobLFNs = {}
    if not retVal["OK"]:
        raise Exception(retVal["Message"])
    success = retVal["Value"]["Successful"]
    jobDict = {}
    for item in list(success):
        jobs = success.pop(item)
        # Note that item may be a jobID or an LFN
        if not item.isdigit():
            #   item is an LFN, hence only one job step (job in BK)
            lfn = item
            job = jobs[0]
            jobLFNs.setdefault(job["Name"], []).append(lfn)
            jobDict.setdefault(job["Name"], job)
        else:
            jobID = item
            # item is a jobID, hence we may have a list of job steps
            if int(jobID) in retVal["Value"]["Failed"]:
                # For some reason sometimes (always?) the jobID is also in Failed
                retVal["Value"]["Failed"].remove(int(jobID))
            jobStr = f"Job {jobID}"
            stepDict = {"Job": {"CPUTIME": 0.0, "ExecTime": 0.0, "NumberOfSteps": len(jobs)}}
            for job in jobs:
                # Group common info into a separate dictionary (and remove it from steps)
                job.pop("DIRACJobId")
                # ... and yes, it is TotalLumonosity ;-)
                for info in (
                    "DIRACVersion",
                    "Location",
                    "Production",
                    "TotalLumonosity",
                    "WNCACHE",
                    "WNCPUHS06",
                    "WNCPUPOWER",
                    "WNMEMORY",
                    "WNMJFHS06",
                    "WNMODEL",
                    "WORKERNODE",
                ):
                    stepDict["Job"][info] = job.pop(info, None)
                # Sum up CPU and wall clock times
                for info in ("CPUTIME", "ExecTime"):
                    stepDict["Job"][info] += job[info]
                step = f"Step {job.pop('Name')}"
                stepDict[step] = job
            success[jobStr] = stepDict

        # When LFNs were given, rearrange the success dictionary in case several files were created by the same step
        if not summary and jobLFNs:
            # Group files produced by the same job
            lfnsByJob = {}
            for name, job in jobDict.items():
                # For each jobID get set of LFNs and job steps
                lfnsByJob.setdefault(job["DIRACJobId"], []).append((jobLFNs[name], job))
            for jobID, lfnJobs in lfnsByJob.items():
                jobStr = f"Job {jobID}"
                # Split job and step information
                stepDict = {jobStr: {}}
                lfnSet = set()
                for lfns, job in lfnJobs:
                    job.pop("DIRACJobId")
                    lfnSet.update(lfns)
                    for info in (
                        "DIRACVersion",
                        "Location",
                        "Production",
                        "TotalLumonosity",
                        "WNCACHE",
                        "WNCPUHS06",
                        "WNCPUPOWER",
                        "WNMEMORY",
                        "WNMJFHS06",
                        "WNMODEL",
                        "WORKERNODE",
                    ):
                        stepDict[jobStr][info] = job.pop(info, None)
                    job[" LFN"] = ",".join(sorted(lfns))
                    stepDict[f"Step {job.pop('Name')}"] = job
                success[",".join(sorted(lfnSet))] = stepDict

    if summary and jobLFNs:
        gLogger.always("List of DIRAC jobs:")
        gLogger.always(",".join(sorted({f"{val['DIRACJobId']}" for val in jobDict.values()})))
    else:
        printDMResult(retVal, empty="File/job does not exist in the Bookkeeping")


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Client plug-in for the RAWIntegrity catalog. This exposes a single method to
add files to the RAW IntegrityDB.

USED at OnLine
"""
from DIRAC import S_OK
from DIRAC.Resources.Catalog.Utilities import checkCatalogArguments
from DIRAC.Resources.Catalog.FileCatalogClientBase import FileCatalogClientBase


class RAWIntegrityClient(FileCatalogClientBase):
    # List of common File Catalog methods implemented by this client
    WRITE_METHODS = FileCatalogClientBase.WRITE_METHODS + ["addFile"]

    def __init__(self, url="", **kwargs):
        self.serverURL = "DataManagement/RAWIntegrity" if not url else url
        super().__init__(self.serverURL, **kwargs)
        self.rawIntegritySrv = self._getRPC()

    def isOK(self):
        """Returns valid."""
        return self.valid

    @checkCatalogArguments
    def addFile(self, lfns):
        failed = {}
        successful = {}
        for lfn, info in lfns.items():
            pfn = str(info["PFN"])
            size = int(info["Size"])
            se = str(info["SE"])
            guid = str(info["GUID"])
            checksum = str(info["Checksum"])
            res = self.rawIntegritySrv.addFile(lfn, pfn, size, se, guid, checksum)
            #       rpc = self._getRPC()
            #       rpc.addFile( lfn, pfn, size, se, guid, checksum )
            if not res["OK"]:
                failed[lfn] = res["Message"]
            else:
                successful[lfn] = True

        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""The ProductionStatusAgent monitors productions for active requests and takes
care to update their status. Initially this is just to handle simulation
requests.

Allowed production status transitions performed by this agent include:

Idle -> ValidatingInput
Idle -> ValidatingOutput

ValidatedOutput -> Completed

ValidatingInput -> RemovingFiles

RemovedFiles -> Completed

Active -> Idle

Testing -> Idle

In addition this also updates request status from Active to Done.

To do: review usage of production API(s) and re-factor into Production Client

AZ 10.14: merged with a part from RequestTrackingAgent to avoid race conditions
"""
import time
import os
import sqlite3
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor, wait

import DIRAC
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.Utilities.TimeUtilities import timeThis
from DIRAC.Interfaces.API.Dirac import Dirac
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations

from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from LHCbDIRAC.Interfaces.API.DiracProduction import DiracProduction
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.ProductionManagementSystem.Client.ProductionRequestClient import ProductionRequestClient

#############################################################################
# The following is used for StandAlone debugging only (outside Agent)
gStandAlone = False  # work in command line without Agent
# gSimulate = gStandAlone and True  # real clients are replaced with simulation
gSimulate = False
gDoRealUpdate = True  # call status updates
gDoRealTracking = True  # update requests progress

KNOWN_TASK_STATES = [
    "Checking",
    "Completed",
    "Created",
    "Matched",
    "Received",
    "Reserved",
    "Rescheduled",
    "Running",
    "Submitted",
    "Waiting",
]


class ProductionStatusAgent(AgentModule):
    """Usual DIRAC agent."""

    def __init__(self, *args, **kwargs):
        """c'tor.

        :param self: self reference
        :param str agentName: name of agent
        :param str loadName: load name of agent
        :param bool baseAgentName: whatever
        :param dict properties: whatever else
        """
        if not gStandAlone:
            AgentModule.__init__(self, *args, **kwargs)
        else:
            self.log = gLogger

        self.dProd = None
        self.dirac = None
        self.prClient = None
        self.tClient = None

        self.simulationTypes = Operations().getValue(
            "Transformations/ExtendableTransfTypes",
            ["MCSimulation", "Simulation"],
        )

        self.allKnownStates = (
            "RemovedFiles",
            "RemovingFiles",
            "ValidatedOutput",
            "ValidatingInput",
            "Testing",
            "Active",
            "Idle",
        )

        self.notify = True
        self.cacheFile = os.path.join(DIRAC.rootPath, "work/ProductionManagement/cache.db")

        # For processing transformations, it can happen that there are some Unused files
        # with which no tasks can be created. The number of such files can be different depending
        # from the module and distribution between centres.
        # So we declare such transformations isIdle() once there is no jobs, no files in other
        # pending states and the number of Unused files was not changed last cyclesTillIdle times
        self.cyclesTillIdle = 1
        self.filesUnused = {}  # <tID: { 'Number': x, 'NotChanged': n }

        self.prMasters = {}  # [ prID: [<subrequests> ...] ]
        self.prSummary = {}
        self.prProds = {}  # <prID>, map production to known request, from _getProductionRequestsProgress
        self.notPrTrans = defaultdict(list)  # transformation without PR, from _getTransformationsState
        self.toUpdate = []

    #############################################################################
    def initialize(self):
        """Sets default values."""
        # shifter
        self.am_setOption("shifterProxy", "ProductionManager")

        if not gStandAlone:
            self.notify = eval(self.am_getOption("NotifyProdManager", "True"))

        # Set the clients
        self.dProd = DiracProduction()
        self.dirac = Dirac()
        if gSimulate:
            raise NotImplementedError()
        self.prClient = ProductionRequestClient()
        self.tClient = TransformationClient()

        return S_OK()

    #############################################################################
    def execute(self):
        """The execution method, track requests progress and implement a part of
        Production SM."""
        updatedT = {}  # updated transformations
        updatedPr = []  # updated production requests (excluding tracking updates)

        # Distinguish between leafs and master requests
        # Masters should not appear in the prodReqSummary and they should have no
        # associated productions.
        self.prMasters = {}  # [ prID: [<subrequests> ...] ]
        self.prSummary = {}
        # { <reqID> :
        #     'type', 'master', 'bkTotal', 'prTotal',  - from _getActiveProductionRequests()
        #     'isDone', 'prods': [ <prodIf> : { 'Used', 'Events' } ] - from __getProductionRequestsProgress
        #     'state' for each production - from _getTransformationsState()
        #     'isIdle', 'isProcIdle' for each 'Active' or 'Idle' production,
        #     'isSimulation' - from _getIdleProductionRequestProductions()
        #     'isFinished' - from _applyProductionRequestsLogic()
        #     'filesTotal' - from _getIdleProductionRequestProductions()
        #     'filesProcessed' - from _getIdleProductionRequestProductions()
        #     'filesUnused' - from _getIdleProductionRequestProductions()
        #     'filesMaxReset' - from _getIdleProductionRequestProductions()
        #     'filesNotProcessed' - from _getIdleProductionRequestProductions()
        #     'inputIDs' - from _getExtraInfo()
        #     'hasEndDate' - from _getExtraInfo()
        #     'hasActiveInput' - from _trackProductionRequests()
        # }
        self.prProds = {}  # <prID>, map production to known request, from _getProductionRequestsProgress

        self.notPrTrans = defaultdict(list)  # transformation without PR, from _getTransformationsState

        self.log.info("******************************")
        self.log.info("Collecting required information")
        self.log.info("******************************")

        result = self._getActiveProductionRequests()
        if not result["OK"]:
            self.log.error("Aborting cycle", result["Message"])
            return S_OK()
        if len(self.prProds) < len(self.prSummary):
            self.log.fatal(
                "Aborting cycle, as the number of sub-requests can't be larger than the number of productions"
            )
            return S_OK()

        self._getTransformationsState()
        result = self._getIdleProductionRequestProductions()
        if not result["OK"]:
            self.log.error("Aborting cycle", result["Message"])
            return S_OK()

        # That is IMPORTANT to do that after we have the transformation status,
        # since Validation can (really???) update BK, rendering MC incomplete
        result = self._trackProductionRequests()  # also updates PR DB
        if not result["OK"]:
            self.log.error("Aborting cycle", result["Message"])
            return S_OK()

        self.log.info("******************************")
        self.log.info("Updating Production Requests and related transformations")
        self.log.info("******************************")

        self._applyProductionRequestsLogic(updatedT, updatedPr)

        self.log.info("******************************")
        self.log.info("Updating Production Request for unrelated transformations (replication, etc.)")
        self.log.info("******************************")

        self._applyOtherTransformationsLogic(updatedT)

        self.log.info("*********")
        self.log.info("Reporting")
        self.log.info("*********")

        if updatedT:
            self.log.info("Transformations updated this cycle:")
            for name, value in updatedT.items():
                self.log.info(f"Transformations {name}: {value['from']} => {value['to']}")

        if updatedPr:
            self.log.info("Production Requests updated to Done status:", f"{', '.join(str(i) for i in updatedPr)}")

        if gDoRealUpdate and not gSimulate:
            self._mailProdManager(updatedT, updatedPr)

        self._cleanFilesUnused()

        if gSimulate:
            self.tClient.animate(self.prClient)

        return S_OK()

    #############################################################################

    @timeThis
    def __getProductionRequestsProgress(self):
        """get known progress for Active requests related productions Failures
        there are critical and can enforce wrong logic."""
        self.log.verbose("Collecting old Production Request Progress...")
        result = self.prClient.getAllProductionProgress()
        if not result["OK"]:
            return S_ERROR(f"Could not retrieve production progress summary: {result['Message']}")
        progressSummary = result["Value"]  # { <prID> : [ <prodId> : { 'Used', 'Events' } ] }

        for prID, summary in self.prSummary.items():
            # Setting it before updating will give grace period before SM ops
            summary["isDone"] = summary["bkTotal"] >= summary["prTotal"]
            summary["prods"] = progressSummary.get(prID, {})
            for tID in summary["prods"]:
                self.prProds[tID] = prID
        self.log.verbose("Done with old Production Request Progress")
        return S_OK()

    @timeThis
    def _getActiveProductionRequests(self):
        """get 'Active' requests.

        Failures there are critical and can enforce wrong logic
        Note: this method can be moved to the service
        """
        self.log.info("Collecting active production requests...")
        result = self.prClient.getProductionRequestList(0, "", "ASC", 0, 0, {"RequestState": "Active"})
        if not result["OK"]:
            return S_ERROR(f"Could not retrieve active production requests: {result['Message']}")
        activeMasters = result["Value"]["Rows"]
        for pr in activeMasters:
            prID = pr["RequestID"]
            if pr["HasSubrequest"]:
                self.prMasters[prID] = []
                result = self.prClient.getProductionRequestList(prID, "", "ASC", 0, 0, {})
                if not result["OK"]:
                    return S_ERROR(f"Could not get subrequests for production request {prID}: {result['Message']}")
                for subPr in result["Value"]["Rows"]:
                    subPrID = subPr["RequestID"]
                    self.prSummary[subPrID] = {
                        "type": pr["RequestType"],
                        "master": prID,
                        "bkTotal": subPr["bkTotal"],
                        "prTotal": subPr["rqTotal"],
                    }
                    self.prMasters[prID].append(subPrID)
            else:
                self.prSummary[prID] = {
                    "type": pr["RequestType"],
                    "master": 0,
                    "bkTotal": pr["bkTotal"],
                    "prTotal": pr["rqTotal"],
                }

        result = self.__getProductionRequestsProgress()
        if not result["OK"]:
            return result

        self.log.info(f"Will work with {len(self.prProds)} productions from {len(self.prSummary)} Active (sub)requests")
        self.log.verbose("Done with collecting Active production requests")
        return S_OK()

    @timeThis
    def __getTransformations(self, status):
        """dev function.

        Get the transformations (print info in the meanwhile)
        """
        res = self.tClient.getTransformationWithStatus(status)
        if not res["OK"]:
            self.log.error("Failed to get transformations", f"{status}: {res['Message']}")
            raise RuntimeError(f"Failed to get {status} transformations: {res['Message']}")
        if not res["Value"]:
            self.log.debug(f"No transformations in {status} status")
            return []
        if len(res["Value"]) > 20:
            self.log.verbose(f"The following number of transformations are in {status} status: {len(res['Value'])}")
        else:
            valOutStr = ", ".join(str(i) for i in res["Value"])
            self.log.verbose(f"The following transformations are in {status} status: {valOutStr}")
        return res["Value"]

    @timeThis
    def _getTransformationsState(self):
        """get Transformations state (set 'Other' for not interesting states)
        failures to get something are not critical since there is no reaction on
        'Other' state."""
        self.log.info("Collecting transformations state...")
        try:
            # We put 'Finished' for both
            tListCompleted = self.__getTransformations("Completed")
            tListArchived = self.__getTransformations("Archived")
            tListFinished = tListCompleted + tListArchived
            for tID in tListFinished:
                prID = self.prProds.get(tID, None)
                if prID:
                    self.prSummary[prID]["prods"][tID]["state"] = "Finished"

            for state in self.allKnownStates:
                tList = self.__getTransformations(state)
                for tID in tList:
                    prID = self.prProds.get(tID, None)
                    if prID:
                        self.prSummary[prID]["prods"][tID]["state"] = state
                    else:
                        self.notPrTrans[state].append(tID)
        except RuntimeError as error:
            self.log.error(error)

        for tID, prID in self.prProds.items():
            if "state" not in self.prSummary[prID]["prods"][tID]:
                self.prSummary[prID]["prods"][tID]["state"] = "Other"

        self.log.verbose("Done with collecting transformations states")

    def _getStatusCountersBulk(self, tableName, statusKey, transformationIDs):
        """Get the counter named ``statusKey`` for many transformations

        Wrapper around ``TransformationClient.getCounters`` that returns a nested
        dictionary of:
        ``{transformationID: {state1: X, state2: Y, ..., "TotalCreated": X+Y+...}}``

        :param str tableName: Table in the TransformationDB to use
        :param str statusKey: Name of the status column to consider
        :param list transformationIDs: List of transformation IDs to inspect
        :returns: ``dict`` of ``dict`` of ``int``
        """
        # This is for compatibility with TransformationClient.getTransformationStats
        # and TransformationClient.getTransformationTaskStats
        if tableName == "TransformationTasks":
            totalColumnName = "TotalCreated"
        elif tableName == "TransformationFiles":
            totalColumnName = "Total"
        else:
            raise NotImplementedError(tableName)
        self.log.verbose("Getting counters from for", f"{len(transformationIDs)} transformations")
        res = self.tClient.getCounters(
            tableName,
            ["TransformationID", statusKey],
            {"TransformationID": transformationIDs},
        )
        if not res["OK"]:
            raise RuntimeError(res)
        # Ensure totalColumnName is always included in the output
        statusDict = defaultdict(lambda: defaultdict(int, **{totalColumnName: 0}))
        for attrDict, count in res["Value"]:
            tID = attrDict["TransformationID"]
            status = attrDict[statusKey]
            statusDict[tID][status] = count
            statusDict[tID][totalColumnName] += count
        # Loop over transformationIDs to ensure all IDs are included in the output
        return {int(tID): dict(statusDict[int(tID)]) for tID in transformationIDs}

    def _isIdleCache(self, transIDs):
        """Get the cache dictionaries that need to be passed to __isIdle

        This method is used with __isIdle to minimise the number of RPC calls required::

          transformations, taskStatuses, fileStatuses = self._isIdleCache(tIDs)
          for tID in tIDs:
            isIdle, isProcIdle, isSimulation = self.__isIdle(
                tID, transformations[tID], taskStatuses[tID], fileStatuses[tID]
            )

        :param list of str tIDs: The IDs to get cached data for
        :returns: transformations, taskStatuses, fileStatuses
        """
        self.log.verbose("Filling _isIdleCache caches for", f"{len(transIDs)} transformations")
        result = self.tClient.getTransformations(condDict={"TransformationID": transIDs}, limit=1000)
        if not result["OK"]:
            self.log.error("Could not get transformations", result["Message"])
            raise RuntimeError("Could not get transformations")
        transformations = {x["TransformationID"]: x for x in result["Value"]}
        taskStatuses = self._getStatusCountersBulk("TransformationTasks", "ExternalStatus", transIDs)
        fileStatuses = self._getStatusCountersBulk("TransformationFiles", "Status", transIDs)
        return transformations, taskStatuses, fileStatuses

    def __isIdle(self, tID, tInfo, tStats, filesStats):
        """Checks if a transformation is idle, is procIdle and either the
        transformation is simulation.

        The ``tInfo``, ``tStats``, ``filesStats`` parameters are passed to improve
        performance by avoiding 3 round trips to the transformation service. See
        ``_isIdleCache`` for more details.

        :param str tID: ID of the transformation to check
        :param dict tInfo: Result of ``TransformationClient.getTransformations``
        :param dict tStats: Result of ``self._getStatusCountersBulk`` for the
                            ``ExternalStatus`` column in the ``TransformationTasks`` table
        :param dict filesStats: Result of ``self._getStatusCountersBulk`` for the
                                ``Status`` column in the ``TransformationFiles`` table
        :returns: ``dict`` of ``dict`` of ``int``
        """
        self.log.debug("Checking if transformation is idle", str(tID))
        if tInfo.get("Type", None) in self.simulationTypes:
            isSimulation = True
            # simulation : go to Idle if
            # only failed and done tasks
            # AND number of tasks created in total == number of tasks submitted
            self.log.verbose("Tasks Stats for %d: %s" % (tID, str(tStats)))
            isIdle = (tStats.get("TotalCreated", 0) > 0) and all(
                tStats.get(status, 0) == 0 for status in KNOWN_TASK_STATES
            )
            isProcIdle = isIdle
        else:
            isSimulation = False
            # other transformation type : go to Idle if
            # 0 assigned files, unused files number was not changing during the last cyclesTillIdle time
            # AND only failed and done tasks
            self.log.debug(f"Files stats: {str(filesStats)}")
            unused = filesStats.get("Unused", 0)
            unusedInherited = filesStats.get("Unused-inherited", 0)
            oldUnused = self.filesUnused.setdefault(tID, {"Number": -1, "NotChanged": 0})
            if oldUnused["Number"] == unused:
                oldUnused["NotChanged"] += 1
            else:
                oldUnused["NotChanged"] = 0
                oldUnused["Number"] = unused
            assigned = filesStats.get("Assigned", 0)
            isProcIdle = (assigned == 0) and ((unused == 0) or (oldUnused["NotChanged"] >= self.cyclesTillIdle))
            if isProcIdle:
                self.log.debug(f"Tasks Stats: {str(tStats)}")
                isProcIdle = all(tStats.get(status, 0) == 0 for status in KNOWN_TASK_STATES)
            isIdle = isProcIdle and (unused == 0) and (unusedInherited == 0)
        return (isIdle, isProcIdle, isSimulation)

    def _getIdleProductionRequestProductions(self):
        """evaluate isIdle and isProcIdle status for all productions we need.

        failures are remembered and are taken into account later
        """
        self.log.verbose("Filling caches for _getIdleProductionRequestProductions...")
        transIDs = [
            str(tID)
            for tID, prID in self.prProds.items()
            if self.prSummary[prID]["prods"][tID]["state"] in ("Active", "Idle")
        ]
        try:
            transformations, taskStatuses, fileStatuses = self._isIdleCache(transIDs)
        except RuntimeError:
            return S_ERROR("Failed to get _isIdleCache in _getIdleProductionRequestProductions")

        self.log.verbose("Checking idle productions...")
        for tID, prID in self.prProds.items():
            tInfo = self.prSummary[prID]["prods"][tID]
            if tInfo["state"] in ("Active", "Idle"):
                isIdle, isProcIdle, isSimulation = self.__isIdle(
                    tID, transformations[tID], taskStatuses[tID], fileStatuses[tID]
                )
                tInfo["isIdle"] = "Yes" if isIdle else "No"
                tInfo["isProcIdle"] = "Yes" if isProcIdle else "No"
                tInfo["isSimulation"] = isSimulation
                tInfo["filesTotal"] = fileStatuses[tID]["Total"]
                tInfo["filesProcessed"] = fileStatuses[tID].get("Processed", 0)
                tInfo["filesUnused"] = fileStatuses[tID].get("Unused", 0)
                tInfo["filesMaxReset"] = fileStatuses[tID].get("MaxReset", 0)
                tInfo["filesNotProcessed"] = fileStatuses[tID].get("NotProcessed", 0) + fileStatuses[tID].get(
                    "Removed", 0
                )
            else:
                tInfo["isIdle"] = "Unknown"
                tInfo["isProcIdle"] = "Unknown"
                tInfo["isSimulation"] = False
                tInfo["filesTotal"] = None
                tInfo["filesProcessed"] = 0
                tInfo["filesUnused"] = 0
                tInfo["filesMaxReset"] = 0
                tInfo["filesNotProcessed"] = 0
        self.log.verbose("Checking idle done")
        return S_OK()

    def _trackProductionRequests(self):
        """contact BK for the current number of processed events failures are
        critical."""
        self.log.info("Updating production requests progress...")

        # Using 10 threads, and waiting for the results before continuing
        futureThreads = []
        with ThreadPoolExecutor(10) as threadPool:
            for tID, prID in self.prProds.items():
                futureThreads.append(threadPool.submit(self._getExtraInfo, tID, prID))
            wait(futureThreads)

        # Update the Production request DB with the number of bookkeeping events
        if self.toUpdate:
            if gDoRealTracking:
                result = self.prClient.updateTrackedProductions(self.toUpdate)
            else:
                result = S_OK()
            if not result["OK"]:
                self.log.error(
                    "Could not send update to the Production Request System", result["Message"]
                )  # that is not critical
            else:
                self.log.verbose(f"The progress of {len(self.toUpdate)} Production Requests is updated")
        self.log.info("Production requests progress update is finished")

        # Get the status of the Analysis Productions input transformations
        inputTransformIDs = set()
        for summary in self.prSummary.values():
            for tID, tInfo in summary["prods"].items():
                inputTransformIDs = inputTransformIDs.union(
                    x for x in tInfo.get("inputIDs", []) if x not in summary["prods"]
                )
        inputTransformStatuses = {}
        if inputTransformIDs:
            retVal = self.tClient.getTransformations(
                condDict={"TransformationID": list(inputTransformIDs)},
                limit=10000,
                columns=["TransformationID", "Status"],
            )
            if not retVal["OK"]:
                self.log.error("Failed to call getTransformations", retVal["Message"])
                return S_ERROR("Too dangerous to continue")
            inputTransformStatuses = {d["TransformationID"]: d["Status"] for d in retVal["Value"]}
        for prID, summary in self.prSummary.items():
            for tID, tInfo in summary["prods"].items():
                if "inputIDs" not in tInfo:
                    continue
                if "hasActiveInput" in tInfo:
                    self.log.info(
                        "hasActiveInput has been overridden for",
                        f"{tID} ({prID}) to be {tInfo['hasActiveInput']}",
                    )
                    continue
                tInfo["hasActiveInput"] = False
                if tInfo["hasEndDate"]:
                    self.log.info(
                        "Skipping hasActiveInput check for",
                        f"{tID} ({prID}) as the input query has an end date",
                    )
                    continue
                for inputID in tInfo["inputIDs"]:
                    if inputID in summary["prods"]:
                        inputState = summary["prods"][inputID]["state"]
                    else:
                        inputState = inputTransformStatuses.get(inputID, "Unknown")
                    if inputState not in ["Archived", "Completed", "Finished", "Cleaned"]:
                        self.log.info(
                            "Marking hasActiveInput=True for",
                            f"{tID} ({prID}) as input {inputID} has status {inputState}",
                        )
                        tInfo["hasActiveInput"] = True

        return S_OK()

    def _getExtraInfo(self, tID, prID):
        tInfo = self.prSummary[prID]["prods"][tID]
        # Get the number of bookkeeping events
        self.log.verbose("Getting BK production progress", "Transformation ID = %d" % tID)
        result = BookkeepingClient().getProductionProducedEvents(tID)
        if result["OK"]:
            nEvents = result["Value"]
            if nEvents and nEvents != tInfo["Events"]:
                self.log.verbose("Updating production %d, with BkEvents %d" % (int(tID), int(nEvents)))
                self.toUpdate.append({"ProductionID": tID, "BkEvents": nEvents})
                tInfo["Events"] = nEvents
        else:
            self.log.error("Progress is not updated", f"{tID} : {result['Message']}")
            return S_ERROR("Too dangerous to continue")
        # Find the transformation ID for the input to each Analysis Production
        if self.prSummary[prID]["type"] == "AnalysisProduction":
            retVal = TransformationClient().getBookkeepingQuery(tID)
            if not retVal["OK"]:
                self.log.error("Failed to call getBookkeepingQuery", f"{tID} {retVal['Message']}")
                return S_ERROR("Too dangerous to continue")
            inputBkQuery = retVal["Value"]
            tInfo["hasEndDate"] = "EndDate" in inputBkQuery
            if not tInfo["hasEndDate"] and inputBkQuery.get("ConfigVersion") in Operations().getValue(
                "AnalysisProductions/ForceActiveInput", []
            ):
                tInfo["hasActiveInput"] = True
            if "ProductionID" in inputBkQuery:
                tInfo["inputIDs"] = [inputBkQuery["ProductionID"]]
            else:
                # FIXME: Just why?!
                if "DataTakingConditions" in inputBkQuery:
                    inputBkQuery["ConditionDescription"] = inputBkQuery["DataTakingConditions"]
                retVal = BookkeepingClient().getProductions(inputBkQuery)
                if not retVal["OK"]:
                    self.log.error("Failed to call getProductions", f"{tID} {retVal['Message']}")
                    return S_ERROR("Too dangerous to continue")
                tInfo["inputIDs"] = [x[0] for x in retVal["Value"]["Records"]]

    def _cleanFilesUnused(self):
        """remove old transformations from filesUnused."""
        oldIDs = []
        for tID in self.filesUnused:
            if tID in self.prProds:
                continue
            if all(tID not in IDs for _status, IDs in self.notPrTrans.items()):
                oldIDs.append(tID)
        for tID in oldIDs:
            del self.filesUnused[tID]

    def __updateTransformationStatus(self, tID, origStatus, status, updatedT):
        """This method updates the transformation status and logs the changes for
        each iteration of the agent.

        Most importantly this method only allows status transitions based on
        what the original status should be.
        """
        self.log.info(f"Changing status for transformation {tID} to {status}")

        if not gDoRealUpdate:
            updatedT[tID] = {"to": status, "from": origStatus}
            return

        result = self.tClient.setTransformationParameter(tID, "Status", status, currentStatus=origStatus)
        if not result["OK"]:
            self.log.error("Failed to update status of transformation", f"{tID} from {origStatus} to {status}")
        else:
            updatedT[tID] = {"to": status, "from": origStatus}

    def _mailProdManager(self, updatedT, updatedPr):
        """Notify the production manager of the changes as productions should be
        manually extended in some cases."""
        if not updatedT and not updatedPr:
            self.log.verbose("No changes this cycle, mail will not be sent")
            return

        if self.notify:
            with sqlite3.connect(self.cacheFile) as conn:
                try:
                    conn.execute(
                        """CREATE TABLE IF NOT EXISTS ProductionStatusAgentCache(
                        production VARCHAR(64) NOT NULL DEFAULT "",
                        from_status VARCHAR(64) NOT NULL DEFAULT "",
                        to_status VARCHAR(64) NOT NULL DEFAULT "",
                        time VARCHAR(64) NOT NULL DEFAULT ""
                       );"""
                    )

                    conn.execute(
                        """CREATE TABLE IF NOT EXISTS ProductionStatusAgentReqCache(
                        prod_requests VARCHAR(64) NOT NULL DEFAULT "",
                        time VARCHAR(64) NOT NULL DEFAULT ""
                       );"""
                    )

                except sqlite3.OperationalError:
                    self.log.error("Could not queue mail")

                for tID, val in updatedT.items():
                    conn.execute(
                        "INSERT INTO ProductionStatusAgentCache (production, from_status, to_status, time)"
                        " VALUES (?, ?, ?, ?)",
                        (tID, val["from"], val["to"], time.asctime()),
                    )

                for prod_request in updatedPr:
                    conn.execute(
                        "INSERT INTO ProductionStatusAgentReqCache (prod_requests, time) VALUES (?, ?)",
                        (prod_request, time.asctime()),
                    )

                conn.commit()

                self.log.info("Mail summary queued for sending")

    def __updateProductionRequestStatus(self, prID, status, updatedPr):
        """This method updates the production request status."""
        self.log.info(f"Marking Production Request {prID} as {status}")

        if not gDoRealUpdate:
            updatedPr.append(prID)
            return

        reqClient = ProductionRequestClient(useCertificates=False, timeout=120)
        result = reqClient.updateProductionRequest(int(prID), {"RequestState": status})
        if not result["OK"]:
            self.log.error(result)
        else:
            updatedPr.append(prID)

    def _applyOtherTransformationsLogic(self, updatedT):
        """animate not Production Requests related transformations failures are not
        critical."""
        self.log.verbose("Updating requests unrelated transformations...")

        self.log.info(
            'Processing %s requests unrelated transformations in "RemovedFiles" state'
            % len(self.notPrTrans["RemovedFiles"])
        )
        for tID in self.notPrTrans["RemovedFiles"]:
            self.__updateTransformationStatus(tID, "RemovedFiles", "Completed", updatedT)

        self.log.info(
            f"Processing {len(self.notPrTrans['Active'])} requests unrelated transformations in \"Active\" state"
        )
        try:
            transformations, taskStatuses, fileStatuses = self._isIdleCache(self.notPrTrans["Active"])
        except RuntimeError:
            self.log.error('Failed to get _isIdleCache for monitoring "Active" requests')
        else:
            for tID in self.notPrTrans["Active"]:
                isIdle, _isProcIdle, _isSimulation = self.__isIdle(
                    tID, transformations[tID], taskStatuses[tID], fileStatuses[tID]
                )
                if isIdle:
                    self.__updateTransformationStatus(tID, "Active", "Idle", updatedT)

        self.log.info(f"Processing {len(self.notPrTrans['Idle'])} requests unrelated transformations in \"Idle\" state")
        try:
            transformations, taskStatuses, fileStatuses = self._isIdleCache(self.notPrTrans["Idle"])
        except RuntimeError:
            self.log.error('Failed to get _isIdleCache for monitoring "Idle" requests')
        else:
            for tID in self.notPrTrans["Idle"]:
                isIdle, _isProcIdle, _isSimulation = self.__isIdle(
                    tID, transformations[tID], taskStatuses[tID], fileStatuses[tID]
                )
                if not isIdle:
                    self.__updateTransformationStatus(tID, "Idle", "Active", updatedT)

        self.log.verbose("Requests unrelated transformations update is finished")

    def _isReallyDone(self, summary):
        """Evaluate 'isDone' from current update cycle."""
        # Other production types rely on the number of bookkeeping events
        bkTotal = sum(t["Events"] for t in summary["prods"].values() if t["Used"])
        return bkTotal >= summary["prTotal"]

    def _producersAreIdle(self, summary):
        """Return True in case all producers (not 'Used') transformations are Idle,
        Finished or not exist."""
        for _tID, tInfo in summary["prods"].items():
            if tInfo["Used"]:
                continue
            if tInfo["isIdle"] != "Yes" and tInfo["state"] != "Finished":
                return False
        return True

    def _producersAreProcIdle(self, summary):
        """Return True in case all producers (not 'Used') transformations are
        procIdle or finished or not exist."""
        for _tID, tInfo in summary["prods"].items():
            if tInfo["Used"]:
                continue
            if tInfo["isProcIdle"] != "Yes" and tInfo["state"] != "Finished":
                return False
        return True

    def _processorsAreProcIdle(self, summary):
        """Return True in case all processors ('Used' or not Sim) transformations
        are procIdle or finished or not exist."""
        for _tID, tInfo in summary["prods"].items():
            if not tInfo["Used"] and tInfo["isSimulation"]:
                continue
            if tInfo["isProcIdle"] != "Yes" and tInfo["state"] != "Finished":
                return False
        return True

    def _mergersAreDone(self, summary):
        """Return True in case all mergers ('Used') transformations are finished or
        not exist."""
        for _tID, tInfo in summary["prods"].items():
            if not tInfo["Used"]:
                continue
            if tInfo["state"] != "Finished":
                return False
        return True

    def _mergersAreProcIdle(self, summary):
        """Return True in case all mergers ('Used') transformations are procIdle or
        finished or not exist."""
        for _tID, tInfo in summary["prods"].items():
            if not tInfo["Used"]:
                continue
            if tInfo["isProcIdle"] != "Yes" and tInfo["state"] != "Finished":
                return False
        return True

    def _requestedMoreThenProduced(self, tID, summary):
        """Check that this transformation has registered less events than it was
        requested."""
        if summary["prods"][tID]["Events"] < summary["prTotal"]:
            self.log.verbose(f" Transformation {tID} has produced less events, asking for extention ")
            return True
        return False

    def _applyProductionRequestsLogic(self, updatedT, updatedPr):
        """Apply known logic to transformations related to production requests
        NOTE: we make decision based on BK statistic collected in the PREVIOUS cycle (except isReallyDone)
              and transformation status in THIS cycle BEFORE the Logic is started
        """
        self.log.verbose("Production Requests logic...")

        for prID, summary in self.prSummary.items():
            countFinished = 0
            for tID, tInfo in summary["prods"].items():
                if tInfo["state"] == "Finished":
                    countFinished += 1
                elif tInfo["state"] == "Idle":
                    self._handleStateIdle(tID, tInfo, summary, updatedT)
                elif tInfo["state"] == "RemovedFiles":
                    self._handleStateRemovedFiles(tID, tInfo, summary, updatedT)
                elif tInfo["state"] == "Active":
                    self._handleStateActive(tID, tInfo, summary, updatedT)
                elif tInfo["state"] == "ValidatedOutput":
                    self._handleStateValidatedOutput(tID, tInfo, summary, updatedT)
                elif tInfo["state"] == "ValidatingInput":
                    self._handleStateValidatingInput(tID, tInfo, summary, updatedT)
                elif tInfo["state"] == "Testing":
                    self._handleStateTesting(tID, tInfo, summary, updatedT)

            summary["isFinished"] = countFinished == len(summary["prods"])
            if (
                summary["isFinished"]
                and not summary["master"]
                and summary["type"] in ["Simulation", "AnalysisProduction"]
            ):
                self.__updateProductionRequestStatus(prID, "Done", updatedPr)

        for masterID, prList in self.prMasters.items():
            countFinished = sum(1 for prID in prList if self.prSummary[prID]["isFinished"])
            if countFinished == len(prList):
                self.__updateProductionRequestStatus(masterID, "Done", updatedPr)

        self.log.verbose("Done with Production Requests logic")

    def _handleStateIdle(self, tID, tInfo, summary, updatedT):
        """Used by _applyProductionRequestsLogic"""
        if tInfo["isIdle"] == "No":
            # 'Idle' && !isIdle() --> 'Active'
            self.__updateTransformationStatus(tID, "Idle", "Active", updatedT)
        elif tInfo["isIdle"] != "Yes":
            return

        if summary["type"] == "Simulation":
            if self._isReallyDone(summary):
                # 'Idle' && isIdle() && isDone for MC logic
                if tInfo["Used"]:  # for standard sim requests, only the merge will go to ValidatingOutput
                    if self._producersAreIdle(summary):
                        self.__updateTransformationStatus(tID, "Idle", "ValidatingOutput", updatedT)
                    # else
                    #  it can happened that MC is !isIdle()
                else:  # for standard sim requests, all but the merge will go to ValidatingInput
                    if self._mergersAreProcIdle(summary):
                        # Note: 'isSimulation' should not be there (it should stay in 'Active')
                        self.__updateTransformationStatus(tID, "Idle", "ValidatingInput", updatedT)
                    # else:
                    #   We wait till mergers finish the job

        elif summary["type"] == "AnalysisProduction":
            if not tInfo["hasActiveInput"]:
                # The number of bookkeeping events is not well defined for Analysis Productions
                allFilesProcessed = (
                    tInfo["filesTotal"] and tInfo["filesTotal"] == tInfo["filesProcessed"] + tInfo["filesNotProcessed"]
                )
                if allFilesProcessed:
                    if tInfo["Used"]:
                        self.__updateTransformationStatus(tID, "Idle", "ValidatingOutput", updatedT)
                    else:
                        self.__updateTransformationStatus(tID, "Idle", "Completed", updatedT)
                elif (
                    tInfo["filesTotal"] == tInfo["filesProcessed"] + tInfo["filesNotProcessed"] + tInfo["filesMaxReset"]
                ):
                    self.log.warn("Transformation has files in MaxReset", f"ID={tID} count={tInfo['filesMaxReset']}")

    def _handleStateRemovedFiles(self, tID, tInfo, summary, updatedT):
        """Used by _applyProductionRequestsLogic"""
        self.__updateTransformationStatus(tID, "RemovedFiles", "Completed", updatedT)

    def _handleStateActive(self, tID, tInfo, summary, updatedT):
        """Used by _applyProductionRequestsLogic"""
        if tInfo["isIdle"] == "Yes":
            if summary["type"] == "Simulation":
                # 'Active' && isIdle() for MC logic
                if tInfo["Used"] or not tInfo["isSimulation"]:
                    # The merger will either wait for MC extention (if !isDone)
                    # or will start validation once producers are isIdle()
                    self.__updateTransformationStatus(tID, "Active", "Idle", updatedT)
                # 'Active' && isIdle() && !Used && isSimulation
                elif self._isReallyDone(summary):
                    if self._mergersAreProcIdle(summary):
                        self.__updateTransformationStatus(tID, "Active", "ValidatingInput", updatedT)
                elif self._processorsAreProcIdle(summary) or self._requestedMoreThenProduced(tID, summary):
                    # we are not done yet, extend production
                    self.__updateTransformationStatus(tID, "Active", "Idle", updatedT)
                # else:
                #  we wait till the situation with mergers is clear
            else:
                # for not MC, use reasonable default
                self.__updateTransformationStatus(tID, "Active", "Idle", updatedT)
        elif summary["type"] == "AnalysisProduction" and not tInfo["hasActiveInput"]:
            if (
                tInfo["filesUnused"]
                and tInfo["filesTotal"]
                == tInfo["filesProcessed"] + tInfo["filesNotProcessed"] + tInfo["filesUnused"] + tInfo["filesMaxReset"]
            ):
                self.__updateTransformationStatus(tID, "Active", "Flush", updatedT)

    def _handleStateValidatedOutput(self, tID, tInfo, summary, updatedT):
        """Used by _applyProductionRequestsLogic"""
        if summary["type"] == "Simulation" and summary["isDone"] and tInfo["Used"]:
            # for standard sim requests, only the merge
            self.__updateTransformationStatus(tID, "ValidatedOutput", "Completed", updatedT)
        elif summary["type"] == "AnalysisProduction" and tInfo["Used"]:
            # isDone only uses the bookkeeping and therefore isn't useful
            self.__updateTransformationStatus(tID, "ValidatedOutput", "Completed", updatedT)
        else:
            self.log.warn(f"Logical bug: transformation {tID} unexpectedly has 'ValidatedOutput'")

    def _handleStateValidatingInput(self, tID, tInfo, summary, updatedT):
        """Used by _applyProductionRequestsLogic"""
        if summary["type"] == "Simulation" and summary["isDone"] and not tInfo["Used"]:
            # for standard sim requests, all but the merge
            self.__updateTransformationStatus(tID, "ValidatingInput", "RemovingFiles", updatedT)
        else:
            self.log.warn(f"Logical bug: transformation {tID} is unexpectedly 'ValidatingInput'")

    def _handleStateTesting(self, tID, tInfo, summary, updatedT):
        """Used by _applyProductionRequestsLogic"""
        try:
            # TODO This should ideally be moved out of the loop
            transformations, taskStatuses, fileStatuses = self._isIdleCache([str(tID)])
        except RuntimeError:
            self.log.error("Failed to get _isIdleCache for", str(tID))
            return
        isIdle, isProcIdle, isSimulation = self.__isIdle(
            tID, transformations[tID], taskStatuses[tID], fileStatuses[tID]
        )
        self.log.verbose("TransID %d, %s, %s, %s" % (tID, isIdle, isProcIdle, isSimulation))
        if isIdle:
            self.__updateTransformationStatus(tID, "Testing", "Idle", updatedT)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Simple extension of ProxyDB for just modifying the message sent.

Ideally, we would not need this code at all, and the message sent should
just be loaded from somewhere else.
"""

from DIRAC import gLogger
from DIRAC.ConfigurationSystem.Client.Helpers import Registry
from DIRAC.FrameworkSystem.DB.ProxyDB import ProxyDB as DIRACProxyDB


class ProxyDB(DIRACProxyDB):
    """Simple extension for just taking care of the message sent."""

    def _notifyProxyAboutToExpire(self, userDN, lTime):
        result = Registry.getUsernameForDN(userDN)
        if not result["OK"]:
            return False
        userName = result["Value"]
        userEMail = Registry.getUserOption(userName, "Email", "")
        if not userEMail:
            gLogger.error("Could not discover user email", userName)
            return False
        daysLeft = int(lTime / 86400)
        msgSubject = "Your proxy uploaded to LHCbDIRAC will expire in %d days" % daysLeft
        msgBody = """\
Dear %s,

  The proxy you uploaded to LHCbDIRAC will expire in approximately %d days. The proxy
  information is:

  DN:    %s

  If you plan on keep using this credentials, please upload a newer proxy to
  LHCbDIRAC by executing (from lxplus.cern.ch, for example):

  $ lhcb-proxy-init --upload

  If you have been issued different certificate, please make sure you have a
  proxy uploaded with that certificate.

Cheers,
 LHCbDIRAC's Proxy Manager
""" % (
            userName,
            daysLeft,
            userDN,
        )
        fromAddr = self._mailFrom
        result = self.__notifClient.sendMail(userEMail, msgSubject, msgBody, fromAddress=fromAddr)
        if not result["OK"]:
            gLogger.error("Could not send email", result["Message"])
            return False
        return True



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
:mod: ValidateRequest

.. module: ValidateRequest

:synopsis: validateRequest operation handler
"""
# # imports
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities import DEncode
from DIRAC.RequestManagementSystem.private.OperationHandlerBase import OperationHandlerBase


class WMSSecureOutputData(OperationHandlerBase):
    """.. class:: ValidateRequest.

    Validate operation handler
    """

    def __init__(self, operation=None, csPath=None):
        """c'tor.

        :param self: self reference
        :param Operation operation: Operation instance
        :param str csPath: CS path for this handler
        """
        OperationHandlerBase.__init__(self, operation, csPath)

    def __call__(self):
        """It expects to find the reqID in operation.Arguments."""
        try:
            decode = DEncode.decode(self.operation.Arguments)
            self.log.debug(decode)
            gLogger.debug("Validating output")
        except ValueError as error:
            self.log.exception(error)
            self.operation.Error = str(error)
            self.operation.Status = "Failed"
            return S_ERROR(str(error))

        self.operation.Status = "Done"
        return S_OK()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""TransformationAgent is and LHCb class just for overwriting some of the DIRAC
methods."""

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.ResourceStatusSystem.Client.ResourceManagementClient import ResourceManagementClient
from DIRAC.TransformationSystem.Agent.TransformationAgent import TransformationAgent as DIRACTransformationAgent

from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

AGENT_NAME = "Transformation/LHCbTransformationAgent"


class TransformationAgent(DIRACTransformationAgent):
    """Extends base class."""

    def initialize(self):
        """LHCb defaults."""
        DIRACTransformationAgent.initialize(self)

        self.pluginLocation = self.am_getOption(
            "PluginLocation", "LHCbDIRAC.TransformationSystem.Agent.TransformationPlugin"
        )
        self.workDirectory = self.am_getWorkDirectory()
        self.debug = self.am_getOption("verbosePlugin", False)

        return S_OK()

    def _getClients(self):
        """returns the clients used in the threads."""
        res = DIRACTransformationAgent._getClients(self)

        threadTransformationClient = TransformationClient()
        threadRMClient = ResourceManagementClient()
        threadBkk = BookkeepingClient()

        res.update(
            {
                "TransformationClient": threadTransformationClient,
                "ResourceManagementClient": threadRMClient,
                "BookkeepingClient": threadBkk,
            }
        )

        return res

    def __generatePluginObject(self, plugin, clients):
        """Generates the plugin object."""
        try:
            plugModule = __import__(self.pluginLocation, globals(), locals(), ["TransformationPlugin"])
        except ImportError as x:
            gLogger.exception(f"{AGENT_NAME}.__generatePluginObject: Failed to import 'TransformationPlugin'", "", x)
            return S_ERROR()
        try:
            oPlugin = getattr(plugModule, "TransformationPlugin")(
                f"{plugin}",
                dataManager=clients["DataManager"],
                transClient=clients["TransformationClient"],
                bkClient=clients["BookkeepingClient"],
                rmClient=clients["ResourceManagementClient"],
            )
        except Exception as x:  # pylint: disable=broad-except
            gLogger.exception(f"{AGENT_NAME}.__generatePluginObject: Failed to create {plugin}().", "", x)
            return S_ERROR()
        oPlugin.workDirectory = self.workDirectory
        oPlugin.pluginCallback = self.pluginCallback
        if self.debug:
            oPlugin.setDebug()
        return S_OK(oPlugin)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Extension of DIRAC Task Manager."""
from DIRAC.TransformationSystem.Client.WorkflowTasks import WorkflowTasks

COMPONENT_NAME = "LHCbTaskManager"


class LHCbWorkflowTasks(WorkflowTasks):
    """A simple LHCb extension to the task manager, for now only used to set the
    runNumber and runMetadata."""

    def _handleInputs(self, oJob, paramsDict):
        """set job inputs (+ metadata)"""
        try:
            if paramsDict["InputData"]:
                self.log.verbose(f"Setting input data to {paramsDict['InputData']}")
                self.log.verbose(f"Setting run number to {str(paramsDict.get('RunNumber'))}")
                oJob.setInputData(paramsDict["InputData"], runNumber=paramsDict.get("RunNumber"))

                try:
                    runMetadata = paramsDict["RunMetadata"]
                    self.log.verbose(f"Setting run metadata information to {str(runMetadata)}")
                    oJob.setRunMetadata(runMetadata)
                except KeyError:
                    pass

        except KeyError:
            self.log.exception("Could not find input data or a run number")
            raise KeyError("Could not found an input data or a run number")

    #############################################################################

    def _handleRest(self, oJob, paramsDict):
        """add as JDL parameters all the other parameters that are not for inputs
        or destination."""

        for paramName, paramValue in paramsDict.items():
            if paramName not in ("InputData", "RunNumber", "RunMetadata", "Site", "TargetSE"):
                if paramValue:
                    self.log.verbose(f"Setting {paramName} to {paramValue}")
                    oJob._addJDLParameter(paramName, paramValue)

    #############################################################################



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

from DIRAC.Core.Base.Script import Script


@Script()
def main():
    Script.parseCommandLine(ignoreErrors=True)

    import DIRAC
    from LHCbDIRAC.Interfaces.API.DiracProduction import DiracProduction

    args = Script.getPositionalArgs()

    diracProd = DiracProduction()

    exitCode = 0
    for prodID in args:
        result = diracProd.getProductionProgress(prodID, printOutput=True)
        if "Message" in result:
            print(f"Listing production summary failed with message:\n{result['Message']}")
            exitCode = 2
        elif not result:
            print("Null result for getProduction() call", prodID)
            exitCode = 2
        else:
            exitCode = 0

    if not args:
        result = diracProd.getProductionProgress(printOutput=True)
        if "Message" in result:
            print(f"Listing production summary failed with message:\n{result['Message']}")
            exitCode = 2
        elif not result:
            print("Null result for getProduction() call")
            exitCode = 2
        else:
            exitCode = 0

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Test class for BookkeepingReport."""
# pylint: disable=missing-docstring, invalid-name, protected-access
import importlib
from itertools import product

import pytest
from unittest.mock import MagicMock

# mocks
from DIRAC.DataManagementSystem.Client.test.mock_DM import dm_mock
from LHCbDIRAC.Workflow.Modules.mock_Commons import (
    prod_id,
    prod_job_id,
    wms_job_id,
    workflowStatus,
    stepStatus,
    step_id,
    step_number,
    step_commons,
    wf_commons,
)
from LHCbDIRAC.BookkeepingSystem.Client.test.mock_BookkeepingClient import bkc_mock

# sut
from LHCbDIRAC.Workflow.Modules.BookkeepingReport import BookkeepingReport


rv_m = importlib.import_module("LHCbDIRAC.Workflow.Modules.ModuleBase")
rv_m.RequestValidator = MagicMock()

mock_pxc = MagicMock()
mock_pxc.return_value = dict()
p_m = importlib.import_module("LHCbDIRAC.Resources.Catalog.PoolXMLFile")
p_m.getOutputType = mock_pxc

bkr = BookkeepingReport(bkClient=bkc_mock, dm=dm_mock)

allCombinations = list(product(wf_commons, step_commons))


@pytest.mark.parametrize("wf_cs, s_cs", allCombinations)
def test_execute(wf_cs, s_cs):
    assert (
        bkr.execute(
            prod_id, prod_job_id, wms_job_id, workflowStatus, stepStatus, wf_cs, s_cs, step_number, step_id, False
        )["OK"]
        is True
    )


@pytest.mark.parametrize(
    "output, outputType, xf_o_r, expected",
    [
        ("/this/is/an/output.txt", "txt", {"/this/is/an/output.txt": {}}, ("{}", "/this/is/an/output.txt")),
        ("/this/is/an/output.hist", "hist", {"/this/is/an/output.txt": {}}, ("Unknown", "/this/is/an/output.hist")),
        ("/this/is/an/output.txt", "txt", {"/this/is/an/OutPut.txt": {}}, ("{}", "/this/is/an/OutPut.txt")),
    ],
)
def test__getFileStatsFromXMLSummary(output, outputType, xf_o_r, expected):
    mock_xfo = MagicMock()
    mock_xfo.outputsEvents = xf_o_r
    bkr.xf_o = mock_xfo

    res = bkr._getFileStatsFromXMLSummary(output, outputType)
    assert res == expected



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Script for scanning the popularity table."""
from DIRAC.Core.Base.Script import Script


@Script()
def main():
    since = 30
    getAllDatasets = False
    Script.registerSwitch("", "Since=", "   Number of days to look for (default: %d)" % since)
    Script.registerSwitch("", "All", "   If used, gets all existing datasets, not only the used ones")
    Script.setUsageMessage(
        "\n".join(
            [
                __doc__,
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile]",
            ]
        )
    )
    Script.parseCommandLine(ignoreErrors=True)
    for switch in Script.getUnprocessedSwitches():
        if switch[0] == "Since":
            try:
                since = int(switch[1])
            except ValueError:
                pass
        elif switch[0] == "All":
            getAllDatasets = True

    from LHCbDIRAC.DataManagementSystem.Client.ScanPopularity import scanPopularity

    scanPopularity(since, getAllDatasets)


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Test XMLErr.py."""
import unittest
import ast
import xml.etree.ElementTree as ET

import LHCbDIRAC.Core.Utilities.XMLErr as XMLErr


class XMLErrTestCase(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def setUp(self):
        self.jobID = 1
        self.prodID = 2
        self.wmsID = 3

    def tearDown(self):
        pass


class TestXMLErr(XMLErrTestCase):
    #########################################
    # Test extractData()

    def test_extractDataWithNoCounter(self):
        rootNoCounter = ET.fromstring("<counters></counters>")
        expected = dict({"Counters": {"ID": {"JobID": 1, "ProductionID": 2, "wmsID": 3}}})
        result = XMLErr.extractData(rootNoCounter, self.jobID, self.prodID, self.wmsID)

        self.assertTrue(result["OK"])
        self.assertEqual(ast.literal_eval(result["Value"]), expected)

    def test_extractDataWithStringValue(self):
        rootStringValue = ET.fromstring('<counters><counter name = "test">"string"</counter></counters>')
        result = XMLErr.extractData(rootStringValue, self.jobID, self.prodID, self.wmsID)

        self.assertFalse(result["OK"])

    def test_extractDataWithFloatValue(self):
        rootFloatValues = ET.fromstring('<counters><counter name = "test">1.5</counter></counters>')
        result = XMLErr.extractData(rootFloatValues, self.jobID, self.prodID, self.wmsID)

        self.assertFalse(result["OK"])

    def test_extractDataNormal(self):
        root = ET.fromstring(
            '<counters><counter name = "test1">1</counter><counter name = "test2">-1</counter></counters>'
        )
        expected = dict({"Counters": {"test1": 1, "test2": -1, "ID": {"JobID": 1, "ProductionID": 2, "wmsID": 3}}})
        result = XMLErr.extractData(root, self.jobID, self.prodID, self.wmsID)

        self.assertTrue(result["OK"])
        self.assertEqual(ast.literal_eval(result["Value"]), expected)


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(XMLErrTestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestXMLErr))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Prepare a file (data.py) which is consumed by Ganga, containing the input
files as resolved in the workflow."""

from DIRAC import S_OK, S_ERROR, gLogger

from LHCbDIRAC.Workflow.Modules.ModuleBase import ModuleBase
from LHCbDIRAC.Core.Utilities.GangaDataFile import GangaDataFile


class CreateDataFile(ModuleBase):
    """CreateDataFile class."""

    def __init__(self, bkClient=None, dm=None):
        """simple init."""

        self.log = gLogger.getSubLogger("CreateDataFile")

        super().__init__(self.log, bkClientIn=bkClient, dm=dm)

        self.gangaFileName = "data.py"
        self.poolXMLCatName = "pool_xml_catalog.xml"
        self.persistency = ""

    def _resolveInputVariables(self):
        """By convention the module parameters are resolved here."""

        super()._resolveInputVariables()
        super()._resolveInputStep()

    def execute(
        self,
        production_id=None,
        prod_job_id=None,
        wms_job_id=None,
        workflowStatus=None,
        stepStatus=None,
        wf_commons=None,
        step_commons=None,
        step_number=None,
        step_id=None,
        gdf=None,
    ):
        """Just calls GangaDataFile with some parameter."""

        try:
            super().execute(
                production_id,
                prod_job_id,
                wms_job_id,
                workflowStatus,
                stepStatus,
                wf_commons,
                step_commons,
                step_number,
                step_id,
            )

            if not self._checkWFAndStepStatus():
                return S_OK()

            self._resolveInputVariables()

            if not gdf:
                gdf = GangaDataFile(fileName=self.gangaFileName, xmlcatalog_file=self.poolXMLCatName, log=self.log)

            if not self.inputDataList:
                return S_OK("No data file generated, because no inputs set")

            gdf.generateDataFile(self.inputDataList, persistency=self.persistency)

            return S_OK()

        except Exception as e:  # pylint:disable=broad-except
            self.log.exception("Failure in CreateDataFile execute module", lException=e)
            return S_ERROR(str(e))

        finally:
            super().finalize()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Test for Production Models"""
import pytest
import yaml
from pathlib import Path

from DIRAC import S_OK
from LHCbDIRAC.ProductionManagementSystem.Utilities.Models import parse_obj
from LHCbDIRAC.ProductionManagementSystem.Utilities.ModelCompatibility import (
    step_to_step_manager_dict,
    runs_to_input_query,
)

EXAMPLE_YAML_FILES = list((Path(__file__).parent / "example_yamls").glob("*.yaml"))


@pytest.mark.parametrize("yaml_path", EXAMPLE_YAML_FILES)
def test_models(yaml_path, monkeypatch):
    monkeypatch.setattr(
        "LHCbDIRAC.ProductionManagementSystem.Utilities.Models.getProxyInfo", lambda: S_OK({"username": "test_user"})
    )

    for pr in yaml.safe_load(yaml_path.read_text()):
        assert parse_obj(pr)


sprucing_expected_step_info = {
    "ApplicationName": "Moore",
    "ApplicationVersion": "v54r15",
    "CONDDB": "",
    "DDDB": "",
    "DQTag": "",
    "OptionFiles": (
        '{"entrypoint":"Hlt2Conf.Sprucing_production:pass_spruce_production",'
        '"extra_options":{"input_raw_format":0.5,"input_type":'
        '"RAW","simulation":false,"data_type":"Upgrade",'
        '"geometry_version":"trunk","conditions_version":"master",'
        '"compression":"ZSTD:1","output_type":"ROOT",'
        '"input_process":"Hlt2","process":"Spruce"},'
        '"extra_args":[]}'
    ),
    "ExtraPackages": "",
    "ProcessingPass": "SprucingPass23",
    "StepName": "Passthrough sprucing",
    "isMulticore": "N",
    "Visible": "Y",
    "mcTCK": "",
    "Usable": "Yes",
    "SystemConfig": "",
}


@pytest.mark.parametrize("yaml_path", EXAMPLE_YAML_FILES)
def test_step_to_step_manager_dict(yaml_path, monkeypatch):
    monkeypatch.setattr(
        "LHCbDIRAC.ProductionManagementSystem.Utilities.Models.getProxyInfo", lambda: S_OK({"username": "test_user"})
    )

    for pr in yaml.safe_load(yaml_path.read_text()):
        p = parse_obj(pr)
        for j, step in enumerate(p.steps, start=1):
            step_info = step_to_step_manager_dict(j, step)
            if step.processing_pass == "SprucingPass23":
                assert step_info["Step"] == sprucing_expected_step_info
                assert step_info["InputFileTypes"] == [
                    {"FileType": "MDF", "Visible": "Y"},
                ]
                assert step_info["OutputFileTypes"] == [
                    {"FileType": "SL.DST", "Visible": "N"},
                    {"FileType": "CHARM.DST", "Visible": "N"},
                    {"FileType": "B2CC.DST", "Visible": "N"},
                    {"FileType": "RD.DST", "Visible": "N"},
                    {"FileType": "BANDQ.DST", "Visible": "N"},
                    {"FileType": "QEE.DST", "Visible": "N"},
                    {"FileType": "B2OC.DST", "Visible": "N"},
                    {"FileType": "BNOC.DST", "Visible": "N"},
                ]


@pytest.mark.parametrize(
    "runs, expected",
    [
        ([], ""),
        ([1], "1"),
        (["34"], "34"),
        (["34", "35"], "34,35"),
        (["34:36"], "34,35,36"),
        (["34:36", "38"], "34,35,36,38"),
        (["34:36", "38:40"], "34,35,36,38,39,40"),
    ],
)
def test_runs_to_input_query(runs, expected):
    assert ",".join(runs_to_input_query(runs)) == expected



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
:mod: PopularityAnalysisAgent

.. module: PopularityAnalysisAgent

:synopsis: The PopularityAnalysis Agent generates the popularity CSV file, sends it for analysis to the
           Yandex data popularity service, and sends the result by email
"""
# imports
import os
import json
import csv
import datetime

from requests import post

# # from DIRAC
from DIRAC import S_OK
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.Utilities.Mail import Mail

from LHCbDIRAC.DataManagementSystem.Client.ScanPopularity import scanPopularity

AGENT_NAME = "DataManagement/PopularityAnalysisAgent"


class FakeException(Exception):
    """Fake exception to drive the error handling."""

    pass


class PopularityAnalysisAgent(AgentModule):
    """
    .. class:: PopularityAnalysisAgent
    """

    def __init__(self, *args, **kwargs):
        """c'tor."""

        super().__init__(*args, **kwargs)

        # Default number of days to analyze
        self.analysisPeriod = 910

        # Top directory to analyze
        self.topDirectory = "/lhcb"

        # address of the DataPopularity server
        self.dataPopularityURL = "http://localhost:5000"

        # Mail address to use as report sender
        self.mailSender = None

        # Mail address to send the report to
        self.mailRecipients = ["lhcb-datamanagement@cern.ch"]

        # Enable sending the email
        self.mailEnabled = True

        # target space saved (in TB)
        self.savedSpaceTarget = 100

        # Min number of replicas
        self.minReplicas = 1

        # Max number of replicas
        self.maxReplicas = 7

        # Placeholder to keep the start date
        self.startDate = None

    def initialize(self):
        """agent initialisation."""

        self.workDirectory = self.am_getOption("WorkDirectory")  # pylint: disable=attribute-defined-outside-init

        self.log.info(f"Working directory: {self.workDirectory}")
        if not os.path.isdir(self.workDirectory):
            os.makedirs(self.workDirectory)

        self.analysisPeriod = self.am_getOption("AnalysisPeriod", self.analysisPeriod)
        self.log.info(f"Analysis period: {self.analysisPeriod}")

        self.topDirectory = self.am_getOption("TopDirectory", self.topDirectory)
        self.log.info(f"Top directory: {self.topDirectory}")

        self.mailSender = self.am_getOption("MailSender", self.mailSender)
        self.log.info(f"Mail sender: {self.mailSender}")

        self.mailRecipients = self.am_getOption("MailRecipients", self.mailRecipients)
        self.log.info(f"Mail recipients: {self.mailRecipients}")

        self.mailEnabled = self.am_getOption("MailEnabled", self.mailEnabled)
        self.log.info(f"Mail enabled: {self.mailEnabled}")

        self.dataPopularityURL = self.am_getOption("DataPopularityURL", self.dataPopularityURL)
        self.log.info(f"Data Popularity URL: {self.dataPopularityURL}")

        self.savedSpaceTarget = self.am_getOption("SavedSpaceTarget", self.savedSpaceTarget)
        self.log.info(f"Data Popularity URL: {self.savedSpaceTarget}")

        self.minReplicas = self.am_getOption("MinReplicas", self.minReplicas)
        self.log.info(f"Min Replicas: {self.minReplicas}")

        self.maxReplicas = self.am_getOption("MaxReplicas", self.maxReplicas)
        self.log.info(f"Max Replicas: {self.maxReplicas}")

        return S_OK()

    def execute(self):
        """Main loop of Popularity agent.

        We first trigger the generation of the csv file which contains all the
        popularity and datasets data.

        We then send it to the Yandex data popularity service, and receive a csv file back.

        We merge the two files into an html human readable file.

        We send the html and the Yandex report by email
        """

        self.startDate = datetime.datetime.utcnow().strftime("%Y-%m-%d_%H-%M")

        popularityFile = os.path.join(self.workDirectory, f"popularity_{self.startDate}.csv")
        csvReportFile = os.path.join(self.workDirectory, f"popularityAnalysis_{self.startDate}.csv")

        scanPopularity(self.analysisPeriod, True, topDirectory=self.topDirectory, csvFile=popularityFile)

        params = json.dumps(
            {"n_tb": self.savedSpaceTarget, "min_replicas": self.minReplicas, "max_replicas": self.maxReplicas}
        )
        errorMail = None
        htmlReportFile = None
        try:
            postResult = post(self.dataPopularityURL, files={"file": open(popularityFile)}, data={"params": params})
            if postResult.status_code != 200:
                errorMail = {"status_code": postResult.status_code, "reason": postResult.reason}
                raise FakeException()
            with open(csvReportFile, "w") as report:
                report.write(postResult.content)

            htmlReportFile = self._generateHtmlReport(popularityFile, csvReportFile)

        except FakeException as e:
            pass
        except Exception as e:  # pylint: disable=broad-except
            self.log.exception("Exception generating the reports", lException=e)
            errorMail = {"status_code": 0, "reason": f"Exception generating the reports {repr(e)}"}
        finally:
            if self.mailEnabled:
                if not errorMail:
                    self._sendReport([csvReportFile, htmlReportFile])
                else:
                    self._sendErrorMail(errorMail)

        return S_OK()

    def _generateHtmlReport(self, popularityFile, inputCsvFile):
        """Generate a report, html formated.

        args:
            popularityFile (str): path to the popularity csv file
            inputCsvFile (str): report of the analysis

        returns:
            path to the html report file
        """

        htmlReportFile = os.path.join(self.workDirectory, f"popularityAnalysis_{self.startDate}.html")

        html_string = """
              <!DOCTYPE html>
               <html>
               <head>
               <meta charset='UTF-8'>
                 <style>
                   table{
                      color:#333;
                      font-family:Helvetica,Arial,sans-serif;
                      min-width:700px;
                      border-collapse:collapse;
                      border-spacing:0
                   }
                   td,th{
                      border:1px solid ;
                      height:30px;
                      transition:all .3s;
                      padding: 15px;
                      padding-top: 0px;
                      padding-bottom: 0px
                   }
                   th{
                      background:#DFDFDF;
                      font-weight:700
                   }
                   td{
                      background:#FAFAFA;
                      text-align:center
                   }
                   tr:nth-child(even) td{background:#F1F1F1}tr:nth-child(odd)
                   td{background:#FEFEFE}tr td:hover{background:#666;color:#FFF}
                 </style>
               </head>
               <body>
                 <table>
                   <tr>
                     <th>Name</th>
                     <th>Size</th>
                     <th>Current Replicas</th>
                     <th>Recommended number of disk replicas</th>
                     <th>Archived</th>
                   </tr>
    """

        archivedDataset = set()

        with open(popularityFile, "rb") as csvfile:
            reader = csv.DictReader(csvfile, delimiter=";")
            for row in reader:
                if row["Nb ArchReps"] >= 1:
                    archivedDataset.add(row["Name"])

        with open(inputCsvFile, "rb") as csvfile:
            reader = csv.DictReader(csvfile, delimiter=",")

            for row in reader:
                row.pop("")
                dsName = row.pop("Name")
                dsSize = row.pop("LFNSize")
                dsReplicas = row.pop("Nb_Replicas")
                dsDecrease = row.pop("DecreaseReplicas")

                line_str = (
                    "<tr>"
                    + "<td>"
                    + "<a title='"
                    + "&#10;".join([":".join(t) for t in row.items()])
                    + "'>"
                    + dsName
                    + "</a></td>"
                    + "<td>"
                    + str(dsSize)
                    + "</td>"
                    + "<td>"
                    + str(dsReplicas)
                    + "</td>"
                    + "<td>"
                    + str(dsDecrease)
                    + "</td>"
                    + "<td>"
                    + str(dsName in archivedDataset)
                    + "</td>"
                    + "</tr>\n"
                )

                html_string += line_str

        html_string += "</table></body></html>"

        with open(htmlReportFile, "w") as ht:
            ht.write(html_string)

        return htmlReportFile

    # pylint: disable=protected-access
    def _sendReport(self, listOfFiles):
        """Send the reports by email.

        args:
            listOfFiles (list): list of files to be sent
        """

        mail = Mail()
        mail._subject = f"Popularity report {self.startDate}"
        mail._message = f" Popularity report {self.startDate}"
        mail._attachments = listOfFiles
        mail._mailAddress = self.mailRecipients
        if self.mailSender:
            mail._fromAddress = self.mailSender
        mail._send()

    # pylint: disable=protected-access
    def _sendErrorMail(self, errorMail):
        """Send the reports by email.

        args:
            errorMail : Error to report
        """

        mail = Mail()
        mail._subject = f"Error popularity report {self.startDate}"
        mail._message = f"Error Popularity report {self.startDate}: {errorMail}"
        mail._mailAddress = self.mailRecipients
        if self.mailSender:
            mail._fromAddress = self.mailSender
        mail._send()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""LHCbDIRAC.ResourceStatusSystem.Policy.DownHillPropagationPolicy.

DownHillPropagationPolicy.__bases__:
  DIRAC.ResourceStatusSystem.PolicySystem.PolicyBase.PolicyBase
"""
from DIRAC.ResourceStatusSystem.PolicySystem.PolicyBase import PolicyBase

# ..............................................................................
#
#
# OBSOLETE CODE. TODO: refactor it !
#
# ..............................................................................


class DownHillPropagationPolicy(PolicyBase):
    """The DownHillPropagationPolicy module is a policy module used to update the
    status of an element, based on how its element in the upper part of the
    hierarchy is behaving in the RSS.

    DownHillPropagationPolicy, given the status of the parent element,
    propagates its status to the child element.
    """

    def evaluate(self):
        """Evaluate policy on "upper" element Status, using args (tuple). The
        status is propagated only when one of the two status is 'Banned'.

        :returns:

        .. code-block:: python

          {
            'Status': Error|Unknown|Active|Probing|Banned,
            'Reason': 'Node/Site status: Active|Probing|Banned'
          }
        """

        commandResult = super().evaluate()
        result = {}

        if commandResult is None:
            result["Status"] = "Error"
            result["Reason"] = "Command evaluation returned None"
            return result

        if not commandResult["OK"]:
            result["Status"] = "Error"
            result["Reason"] = commandResult["Message"]
            return result

        commandResult = commandResult["Value"]

        if commandResult is None:
            result["Status"] = "Unknown"
            result["Reason"] = "No values to take a decission"
            return result

        result["Status"] = commandResult
        result["Reason"] = f"DownHill propagated status: {commandResult}"
        return result
